import time
import unittest
from unittest import mock
from unittest.mock import call

from betfairlightweight import BetfairError

from flumine import config
from flumine.clients.clients import ExchangeType
from flumine.exceptions import OrderExecutionError
from flumine.execution.baseexecution import (
    MAX_SESSION_AGE,
    BET_ID_START,
    BaseExecution,
    OrderPackageType,
)
from flumine.execution.betfairexecution import BetfairExecution
from flumine.execution.simulatedexecution import SimulatedExecution


class BaseExecutionTest(unittest.TestCase):
    def setUp(self) -> None:
        self.mock_flumine = mock.Mock()
        self.execution = BaseExecution(self.mock_flumine, max_workers=2)

    def test_init(self):
        self.assertEqual(MAX_SESSION_AGE, 200)
        self.assertEqual(BET_ID_START, 100000000000)
        self.assertEqual(self.execution.flumine, self.mock_flumine)
        self.assertEqual(self.execution._max_workers, 2)
        self.assertIsNotNone(self.execution._thread_pool)
        self.assertIsNone(self.execution.EXCHANGE)
        self.assertEqual(self.execution._bet_id, 100000000000)
        self.assertEqual(self.execution._sessions, [])
        self.assertEqual(self.execution._sessions_created, 0)

    @mock.patch("flumine.execution.baseexecution.BaseExecution._get_http_session")
    @mock.patch("flumine.execution.baseexecution.BaseExecution.execute_place")
    def test_handler_place(self, mock_execute_place, mock__get_http_session):
        mock_execute_place.__name__ = "execute_place"
        mock_order_package = mock.Mock(elapsed_seconds=1)
        mock_order_package.package_type = OrderPackageType.PLACE
        mock_thread_pool = mock.Mock(_threads=())
        self.execution._thread_pool = mock_thread_pool
        self.execution.handler(mock_order_package)
        mock_thread_pool.submit.assert_called_with(
            mock_execute_place, mock_order_package, mock__get_http_session()
        )
        mock__get_http_session.assert_called_with()

    @mock.patch("flumine.execution.baseexecution.BaseExecution._get_http_session")
    @mock.patch("flumine.execution.baseexecution.BaseExecution.execute_cancel")
    def test_handler_cancel(self, mock_execute_cancel, mock__get_http_session):
        mock_execute_cancel.__name__ = "execute_cancel"
        mock_order_package = mock.Mock(elapsed_seconds=1)
        mock_order_package.package_type = OrderPackageType.PLACE.CANCEL
        mock_thread_pool = mock.Mock(_threads=())
        self.execution._thread_pool = mock_thread_pool
        self.execution.handler(mock_order_package)
        mock_thread_pool.submit.assert_called_with(
            mock_execute_cancel, mock_order_package, mock__get_http_session()
        )
        mock__get_http_session.assert_called_with()

    @mock.patch("flumine.execution.baseexecution.BaseExecution._get_http_session")
    @mock.patch("flumine.execution.baseexecution.BaseExecution.execute_replace")
    def test_handler_replace(self, mock_execute_replace, mock__get_http_session):
        mock_execute_replace.__name__ = "execute_replace"
        mock_order_package = mock.Mock(elapsed_seconds=1)
        mock_order_package.package_type = OrderPackageType.REPLACE
        mock_thread_pool = mock.Mock(_threads=())
        self.execution._thread_pool = mock_thread_pool
        self.execution.handler(mock_order_package)
        mock_thread_pool.submit.assert_called_with(
            mock_execute_replace, mock_order_package, mock__get_http_session()
        )
        mock__get_http_session.assert_called_with()

    @mock.patch("flumine.execution.baseexecution.BaseExecution._get_http_session")
    @mock.patch("flumine.execution.baseexecution.BaseExecution.execute_update")
    def test_handler_update(self, mock_execute_update, mock__get_http_session):
        mock_execute_update.__name__ = "execute_update"
        mock_order_package = mock.Mock(elapsed_seconds=1)
        mock_order_package.package_type = OrderPackageType.UPDATE
        mock_thread_pool = mock.Mock(_threads=())
        self.execution._thread_pool = mock_thread_pool
        self.execution.handler(mock_order_package)
        mock_thread_pool.submit.assert_called_with(
            mock_execute_update, mock_order_package, mock__get_http_session()
        )
        mock__get_http_session.assert_called_with()

    def test_handler_unknown(self):
        mock_order_package = mock.Mock()
        mock_order_package.package_type = "DELETE"
        with self.assertRaises(NotImplementedError):
            self.execution.handler(mock_order_package)

    def test_execute_place(self):
        with self.assertRaises(NotImplementedError):
            self.execution.execute_place(None, None)

    def test_execute_cancel(self):
        with self.assertRaises(NotImplementedError):
            self.execution.execute_place(None, None)

    def test_execute_update(self):
        with self.assertRaises(NotImplementedError):
            self.execution.execute_place(None, None)

    def test_execute_replace(self):
        with self.assertRaises(NotImplementedError):
            self.execution.execute_place(None, None)

    @mock.patch("flumine.execution.baseexecution.BaseExecution._create_new_session")
    def test__get_http_session(self, mock__create_new_session):
        mock_session_one = mock.Mock(time_returned=time.time())
        mock_session_two = mock.Mock(time_returned=time.time())
        self.execution._sessions = [mock_session_one, mock_session_two]
        self.assertEqual(self.execution._get_http_session(), mock_session_one)
        self.assertEqual(self.execution._get_http_session(), mock_session_two)
        self.assertEqual(self.execution._get_http_session(), mock__create_new_session())

    @mock.patch("flumine.execution.baseexecution.BaseExecution._return_http_session")
    @mock.patch("flumine.execution.baseexecution.BaseExecution._create_new_session")
    def test__get_http_session_stale(
        self, mock__create_new_session, mock__return_http_session
    ):
        mock_session = mock.Mock(time_returned=123)
        self.execution._sessions = [mock_session]
        self.assertEqual(self.execution._get_http_session(), mock__create_new_session())
        mock__return_http_session.assert_called_with(mock_session, err=True)

    def test__create_new_session(self):
        session = self.execution._create_new_session()
        self.assertIsNotNone(session.time_created)
        self.assertIsNotNone(session.time_returned)

    def test__return_http_session(self):
        mock_session = mock.Mock()
        self.execution._return_http_session(mock_session)
        self.assertEqual(self.execution._sessions, [mock_session])
        self.execution._return_http_session(mock_session)
        self.execution._return_http_session(mock_session)
        self.assertEqual(self.execution._sessions, [mock_session, mock_session])
        self.assertGreater(mock_session.time_returned, 0)

    def test__return_http_session_returned(self):
        self.execution._sessions = [1, 2]
        mock_session = mock.Mock()
        self.execution._return_http_session(mock_session)

    def test__return_http_session_err_close(self):
        mock_session = mock.Mock()
        self.execution._return_http_session(mock_session, err=True)

    @mock.patch("flumine.execution.baseexecution.OrderEvent")
    def test__order_logger_place(self, mock_order_event):
        mock_order = mock.Mock(async_=False)
        mock_instruction_report = mock.Mock()
        self.execution._order_logger(
            mock_order, mock_instruction_report, OrderPackageType.PLACE
        )
        self.assertEqual(mock_order.bet_id, mock_instruction_report.bet_id)
        mock_order.responses.placed.assert_called_with(mock_instruction_report, dt=True)
        self.mock_flumine.log_control.assert_called_with(mock_order_event(mock_order))

    @mock.patch("flumine.execution.baseexecution.OrderEvent")
    def test__order_logger_place_async(self, mock_order_event):
        mock_order = mock.Mock(async_=True, simulated=False)
        mock_instruction_report = mock.Mock()
        self.execution._order_logger(
            mock_order, mock_instruction_report, OrderPackageType.PLACE
        )
        self.assertEqual(mock_order.bet_id, mock_instruction_report.bet_id)
        mock_order.responses.placed.assert_called_with(
            mock_instruction_report, dt=False
        )
        self.mock_flumine.log_control.assert_called_with(mock_order_event(mock_order))

    @mock.patch("flumine.execution.baseexecution.OrderEvent")
    def test__order_logger_place_async_simulated(self, mock_order_event):
        mock_order = mock.Mock(async_=True, simulated=True)
        mock_instruction_report = mock.Mock()
        self.execution._order_logger(
            mock_order, mock_instruction_report, OrderPackageType.PLACE
        )
        self.assertEqual(mock_order.bet_id, mock_instruction_report.bet_id)
        mock_order.responses.placed.assert_called_with(mock_instruction_report, dt=True)
        self.mock_flumine.log_control.assert_called_with(mock_order_event(mock_order))

    def test__order_logger_place_no_bet_id(self):
        mock_order = mock.Mock(bet_id=123, async_=False)
        mock_instruction_report = mock.Mock(bet_id=None)
        self.execution._order_logger(
            mock_order, mock_instruction_report, OrderPackageType.PLACE
        )
        self.assertEqual(mock_order.bet_id, 123)
        mock_order.responses.placed.assert_called_with(mock_instruction_report, dt=True)
        self.mock_flumine.log_control.assert_not_called()

    def test__order_logger_cancel(self):
        mock_order = mock.Mock()
        mock_instruction_report = mock.Mock()
        self.execution._order_logger(
            mock_order, mock_instruction_report, OrderPackageType.CANCEL
        )
        mock_order.responses.cancelled.assert_called_with(mock_instruction_report)

    def test__order_logger_update(self):
        mock_order = mock.Mock()
        mock_instruction_report = mock.Mock()
        self.execution._order_logger(
            mock_order, mock_instruction_report, OrderPackageType.UPDATE
        )
        mock_order.responses.updated.assert_called_with(mock_instruction_report)

    @mock.patch("flumine.execution.baseexecution.OrderEvent")
    def test__order_logger_replace(self, mock_order_event):
        mock_order = mock.Mock()
        mock_instruction_report = mock.Mock()
        self.execution._order_logger(
            mock_order, mock_instruction_report, OrderPackageType.REPLACE
        )
        self.assertEqual(mock_order.bet_id, mock_instruction_report.bet_id)
        mock_order.responses.placed.assert_called_with(mock_instruction_report)
        self.mock_flumine.log_control.assert_called_with(mock_order_event(mock_order))

    def test_shutdown(self):
        self.execution.shutdown()
        self.assertTrue(self.execution._thread_pool._shutdown)


class BetfairExecutionTest(unittest.TestCase):
    def setUp(self) -> None:
        self.mock_flumine = mock.Mock()
        self.execution = BetfairExecution(self.mock_flumine)

    def test_init(self):
        self.assertEqual(self.execution.EXCHANGE, ExchangeType.BETFAIR)

    @mock.patch("flumine.execution.betfairexecution.BetfairExecution._order_logger")
    @mock.patch("flumine.execution.betfairexecution.BetfairExecution.place")
    @mock.patch("flumine.execution.betfairexecution.BetfairExecution._execution_helper")
    def test_execute_place_success(
        self, mock__execution_helper, mock_place, mock__order_logger
    ):
        mock_session = mock.Mock()
        mock_order = mock.Mock()
        mock_order.trade.__enter__ = mock.Mock()
        mock_order.trade.__exit__ = mock.Mock()
        mock_order_package = mock.MagicMock()
        mock_order_package.__len__.return_value = 1
        mock_order_package.__iter__ = mock.Mock(return_value=iter([mock_order]))
        mock_order_package.info = {}
        mock_report = mock.Mock()
        mock_instruction_report = mock.Mock()
        mock_instruction_report.status = "SUCCESS"
        mock_report.place_instruction_reports = [mock_instruction_report]
        mock__execution_helper.return_value = mock_report
        self.execution.execute_place(mock_order_package, mock_session)
        mock__execution_helper.assert_called_with(
            mock_place, mock_order_package, mock_session
        )
        mock__order_logger.assert_called_with(
            mock_order, mock_instruction_report, OrderPackageType.PLACE
        )
        mock_order.executable.assert_called_with()
        mock_order.trade.__enter__.assert_called_with()
        mock_order.trade.__exit__.assert_called_with(None, None, None)
        mock_order_package.client.add_transaction.assert_called_with(1)

    @mock.patch("flumine.execution.betfairexecution.BetfairExecution._order_logger")
    @mock.patch("flumine.execution.betfairexecution.BetfairExecution.place")
    @mock.patch("flumine.execution.betfairexecution.BetfairExecution._execution_helper")
    def test_execute_place_success_pending(
        self, mock__execution_helper, mock_place, mock__order_logger
    ):
        mock_session = mock.Mock()
        mock_order = mock.Mock()
        mock_order.trade.__enter__ = mock.Mock()
        mock_order.trade.__exit__ = mock.Mock()
        mock_order_package = mock.MagicMock()
        mock_order_package.__len__.return_value = 1
        mock_order_package.__iter__ = mock.Mock(return_value=iter([mock_order]))
        mock_order_package.info = {}
        mock_report = mock.Mock()
        mock_instruction_report = mock.Mock(status="SUCCESS", order_status="PENDING")
        mock_report.place_instruction_reports = [mock_instruction_report]
        mock__execution_helper.return_value = mock_report
        self.execution.execute_place(mock_order_package, mock_session)
        mock__execution_helper.assert_called_with(
            mock_place, mock_order_package, mock_session
        )
        mock__order_logger.assert_called_with(
            mock_order, mock_instruction_report, OrderPackageType.PLACE
        )
        mock_order.executable.assert_not_called()
        mock_order.trade.__enter__.assert_called_with()
        mock_order.trade.__exit__.assert_called_with(None, None, None)
        mock_order_package.client.add_transaction.assert_called_with(1)

    @mock.patch("flumine.execution.betfairexecution.BetfairExecution._order_logger")
    @mock.patch("flumine.execution.betfairexecution.BetfairExecution.place")
    @mock.patch("flumine.execution.betfairexecution.BetfairExecution._execution_helper")
    def test_execute_place_failure(
        self, mock__execution_helper, mock_place, mock__order_logger
    ):
        mock_session = mock.Mock()
        mock_order = mock.Mock()
        mock_order.trade.__enter__ = mock.Mock()
        mock_order.trade.__exit__ = mock.Mock()
        mock_order_package = mock.MagicMock()
        mock_order_package.__len__.return_value = 1
        mock_order_package.__iter__ = mock.Mock(return_value=iter([mock_order]))
        mock_order_package.info = {}
        mock_report = mock.Mock()
        mock_instruction_report = mock.Mock()
        mock_instruction_report.status = "FAILURE"
        mock_report.place_instruction_reports = [mock_instruction_report]
        mock__execution_helper.return_value = mock_report
        self.execution.execute_place(mock_order_package, mock_session)
        mock__execution_helper.assert_called_with(
            mock_place, mock_order_package, mock_session
        )
        mock__order_logger.assert_called_with(
            mock_order, mock_instruction_report, OrderPackageType.PLACE
        )
        mock_order.execution_complete.assert_called_with()
        mock_order.trade.__enter__.assert_called_with()
        mock_order.trade.__exit__.assert_called_with(None, None, None)
        mock_order_package.client.add_transaction.assert_called_with(1)

    @mock.patch("flumine.execution.betfairexecution.BetfairExecution._order_logger")
    @mock.patch("flumine.execution.betfairexecution.BetfairExecution.place")
    @mock.patch("flumine.execution.betfairexecution.BetfairExecution._execution_helper")
    def test_execute_place_timeout(
        self, mock__execution_helper, mock_place, mock__order_logger
    ):
        mock_session = mock.Mock()
        mock_order = mock.Mock()
        mock_order.trade.__enter__ = mock.Mock()
        mock_order.trade.__exit__ = mock.Mock()
        mock_order_package = mock.MagicMock()
        mock_order_package.__len__.return_value = 1
        mock_order_package.__iter__ = mock.Mock(return_value=iter([mock_order]))
        mock_order_package.info = {}
        mock_report = mock.Mock()
        mock_instruction_report = mock.Mock()
        mock_instruction_report.status = "TIMEOUT"
        mock_report.place_instruction_reports = [mock_instruction_report]
        mock__execution_helper.return_value = mock_report
        self.execution.execute_place(mock_order_package, mock_session)
        mock__execution_helper.assert_called_with(
            mock_place, mock_order_package, mock_session
        )
        mock__order_logger.assert_called_with(
            mock_order, mock_instruction_report, OrderPackageType.PLACE
        )
        mock_order.trade.__enter__.assert_called_with()
        mock_order.trade.__exit__.assert_called_with(None, None, None)
        mock_order_package.client.add_transaction.assert_called_with(1)

    @mock.patch("flumine.execution.betfairexecution.BetfairExecution.place")
    @mock.patch("flumine.execution.betfairexecution.BetfairExecution._execution_helper")
    def test_execute_place_error(self, mock__execution_helper, mock_place):
        mock_session = mock.Mock()
        mock_order_package = mock.Mock()
        mock_order_package.info = {}
        mock__execution_helper.return_value = None
        self.execution.execute_place(mock_order_package, mock_session)
        mock__execution_helper.assert_called_with(
            mock_place, mock_order_package, mock_session
        )

    def test_place(self):
        mock_session = mock.Mock()
        mock_order_package = mock.Mock()
        self.assertEqual(
            self.execution.place(mock_order_package, mock_session),
            mock_order_package.client.betting_client.betting.place_orders(),
        )

    @mock.patch("flumine.execution.betfairexecution.BetfairExecution._order_logger")
    @mock.patch("flumine.execution.betfairexecution.BetfairExecution.cancel")
    @mock.patch("flumine.execution.betfairexecution.BetfairExecution._execution_helper")
    def test_execute_cancel_success(
        self, mock__execution_helper, mock_cancel, mock__order_logger
    ):
        mock_session = mock.Mock()
        mock_order = mock.Mock()
        mock_order.trade.__enter__ = mock.Mock()
        mock_order.trade.__exit__ = mock.Mock()
        mock_order.bet_id = 123
        mock_order.size_remaining = 2
        mock_order_package = mock.Mock()
        mock_order_package.__iter__ = mock.Mock(return_value=iter([mock_order]))
        mock_order_package.info = {}
        mock_report = mock.Mock()
        mock_instruction_report = mock.Mock()
        mock_instruction_report.status = "SUCCESS"
        mock_instruction_report.instruction.bet_id = 123
        mock_instruction_report.size_cancelled = 2
        mock_report.cancel_instruction_reports = [mock_instruction_report]
        mock__execution_helper.return_value = mock_report
        self.execution.execute_cancel(mock_order_package, mock_session)
        mock__execution_helper.assert_called_with(
            mock_cancel, mock_order_package, mock_session
        )
        mock__order_logger.assert_called_with(
            mock_order, mock_instruction_report, OrderPackageType.CANCEL
        )
        mock_order.execution_complete.assert_called_with()
        mock_order.trade.__enter__.assert_called_with()
        mock_order.trade.__exit__.assert_called_with(None, None, None)

    @mock.patch("flumine.execution.betfairexecution.BetfairExecution._order_logger")
    @mock.patch("flumine.execution.betfairexecution.BetfairExecution.cancel")
    @mock.patch("flumine.execution.betfairexecution.BetfairExecution._execution_helper")
    def test_execute_cancel_success_race(
        self, mock__execution_helper, mock_cancel, mock__order_logger
    ):
        mock_session = mock.Mock()
        mock_order = mock.Mock()
        mock_order.trade.__enter__ = mock.Mock()
        mock_order.trade.__exit__ = mock.Mock()
        mock_order.bet_id = 123
        mock_order.size_remaining = 0
        mock_order_package = mock.Mock()
        mock_order_package.__iter__ = mock.Mock(return_value=iter([mock_order]))
        mock_order_package.info = {}
        mock_report = mock.Mock()
        mock_instruction_report = mock.Mock()
        mock_instruction_report.status = "SUCCESS"
        mock_instruction_report.instruction.bet_id = 123
        mock_instruction_report.size_cancelled = 2
        mock_report.cancel_instruction_reports = [mock_instruction_report]
        mock__execution_helper.return_value = mock_report
        self.execution.execute_cancel(mock_order_package, mock_session)
        mock__execution_helper.assert_called_with(
            mock_cancel, mock_order_package, mock_session
        )
        mock__order_logger.assert_called_with(
            mock_order, mock_instruction_report, OrderPackageType.CANCEL
        )
        mock_order.execution_complete.assert_called_with()
        mock_order.trade.__enter__.assert_called_with()
        mock_order.trade.__exit__.assert_called_with(None, None, None)

    @mock.patch("flumine.execution.betfairexecution.BetfairExecution._order_logger")
    @mock.patch("flumine.execution.betfairexecution.BetfairExecution.cancel")
    @mock.patch("flumine.execution.betfairexecution.BetfairExecution._execution_helper")
    def test_execute_cancel_success_partial(
        self, mock__execution_helper, mock_cancel, mock__order_logger
    ):
        mock_session = mock.Mock()
        mock_order = mock.Mock()
        mock_order.trade.__enter__ = mock.Mock()
        mock_order.trade.__exit__ = mock.Mock()
        mock_order.bet_id = 123
        mock_order.size_remaining = 2
        mock_order_package = mock.Mock()
        mock_order_package.__iter__ = mock.Mock(return_value=iter([mock_order]))
        mock_order_package.info = {}
        mock_report = mock.Mock()
        mock_instruction_report = mock.Mock()
        mock_instruction_report.status = "SUCCESS"
        mock_instruction_report.instruction.bet_id = 123
        mock_instruction_report.size_cancelled = 1
        mock_report.cancel_instruction_reports = [mock_instruction_report]
        mock__execution_helper.return_value = mock_report
        self.execution.execute_cancel(mock_order_package, mock_session)
        mock__execution_helper.assert_called_with(
            mock_cancel, mock_order_package, mock_session
        )
        mock__order_logger.assert_called_with(
            mock_order, mock_instruction_report, OrderPackageType.CANCEL
        )
        mock_order.executable.assert_called_with()
        mock_order.trade.__enter__.assert_called_with()
        mock_order.trade.__exit__.assert_called_with(None, None, None)

    @mock.patch("flumine.execution.betfairexecution.BetfairExecution._order_logger")
    @mock.patch("flumine.execution.betfairexecution.BetfairExecution.cancel")
    @mock.patch("flumine.execution.betfairexecution.BetfairExecution._execution_helper")
    def test_execute_cancel_failure(
        self, mock__execution_helper, mock_cancel, mock__order_logger
    ):
        mock_session = mock.Mock()
        mock_order = mock.Mock()
        mock_order.trade.__enter__ = mock.Mock()
        mock_order.trade.__exit__ = mock.Mock()
        mock_order.bet_id = 123
        mock_order_package = mock.Mock()
        mock_order_package.__iter__ = mock.Mock(return_value=iter([mock_order]))
        mock_order_package.info = {}
        mock_report = mock.Mock()
        mock_instruction_report = mock.Mock()
        mock_instruction_report.status = "FAILURE"
        mock_instruction_report.instruction.bet_id = 123
        mock_report.cancel_instruction_reports = [mock_instruction_report]
        mock__execution_helper.return_value = mock_report
        self.execution.execute_cancel(mock_order_package, mock_session)
        mock__execution_helper.assert_called_with(
            mock_cancel, mock_order_package, mock_session
        )
        mock__order_logger.assert_called_with(
            mock_order, mock_instruction_report, OrderPackageType.CANCEL
        )
        mock_order.executable.assert_called_with()
        mock_order.trade.__enter__.assert_called_with()
        mock_order.trade.__exit__.assert_called_with(None, None, None)
        mock_order_package.client.add_transaction.assert_called_with(1, failed=True)

    @mock.patch("flumine.execution.betfairexecution.BetfairExecution._order_logger")
    @mock.patch("flumine.execution.betfairexecution.BetfairExecution.cancel")
    @mock.patch("flumine.execution.betfairexecution.BetfairExecution._execution_helper")
    def test_execute_cancel_failure_lapsed(
        self, mock__execution_helper, mock_cancel, mock__order_logger
    ):
        mock_session = mock.Mock()
        mock_order = mock.Mock(bet_id=123)
        mock_order.trade.__enter__ = mock.Mock()
        mock_order.trade.__exit__ = mock.Mock()
        mock_order_package = mock.Mock(info={})
        mock_order_package.__iter__ = mock.Mock(return_value=iter([mock_order]))
        mock_report = mock.Mock()
        mock_instruction_report = mock.Mock(
            status="FAILURE", error_code="BET_TAKEN_OR_LAPSED"
        )
        mock_instruction_report.instruction.bet_id = 123
        mock_report.cancel_instruction_reports = [mock_instruction_report]
        mock__execution_helper.return_value = mock_report
        self.execution.execute_cancel(mock_order_package, mock_session)
        mock__execution_helper.assert_called_with(
            mock_cancel, mock_order_package, mock_session
        )
        mock__order_logger.assert_called_with(
            mock_order, mock_instruction_report, OrderPackageType.CANCEL
        )
        mock_order.execution_complete.assert_called_with()
        mock_order.trade.__enter__.assert_called_with()
        mock_order.trade.__exit__.assert_called_with(None, None, None)
        mock_order_package.client.add_transaction.assert_called_with(1, failed=True)

    @mock.patch("flumine.execution.betfairexecution.BetfairExecution._order_logger")
    @mock.patch("flumine.execution.betfairexecution.BetfairExecution.cancel")
    @mock.patch("flumine.execution.betfairexecution.BetfairExecution._execution_helper")
    def test_execute_cancel_timeout(
        self, mock__execution_helper, mock_cancel, mock__order_logger
    ):
        mock_session = mock.Mock()
        mock_order = mock.Mock()
        mock_order.trade.__enter__ = mock.Mock()
        mock_order.trade.__exit__ = mock.Mock()
        mock_order.bet_id = 123
        mock_order_package = mock.Mock()
        mock_order_package.__iter__ = mock.Mock(return_value=iter([mock_order]))
        mock_order_package.info = {}
        mock_report = mock.Mock()
        mock_instruction_report = mock.Mock()
        mock_instruction_report.status = "TIMEOUT"
        mock_instruction_report.instruction.bet_id = 123
        mock_report.cancel_instruction_reports = [mock_instruction_report]
        mock__execution_helper.return_value = mock_report
        self.execution.execute_cancel(mock_order_package, mock_session)
        mock__execution_helper.assert_called_with(
            mock_cancel, mock_order_package, mock_session
        )
        mock__order_logger.assert_called_with(
            mock_order, mock_instruction_report, OrderPackageType.CANCEL
        )
        mock_order.executable.assert_called_with()
        mock_order.trade.__enter__.assert_called_with()
        mock_order.trade.__exit__.assert_called_with(None, None, None)

    @mock.patch("flumine.execution.betfairexecution.BetfairExecution._order_logger")
    @mock.patch("flumine.execution.betfairexecution.BetfairExecution.cancel")
    @mock.patch("flumine.execution.betfairexecution.BetfairExecution._execution_helper")
    def test_execute_cancel_not_returned(
        self, mock__execution_helper, mock_cancel, mock__order_logger
    ):
        mock_session = mock.Mock()
        mock_order = mock.Mock()
        mock_order.trade.__enter__ = mock.Mock()
        mock_order.trade.__exit__ = mock.Mock()
        mock_order.bet_id = 123
        mock_order_package = mock.Mock()
        mock_order_package.__iter__ = mock.Mock(return_value=iter([mock_order]))
        mock_order_package.info = {}
        mock_report = mock.Mock()
        mock_report.cancel_instruction_reports = []
        mock__execution_helper.return_value = mock_report
        self.execution.execute_cancel(mock_order_package, mock_session)
        mock__execution_helper.assert_called_with(
            mock_cancel, mock_order_package, mock_session
        )
        mock_order.executable.assert_called_with()
        mock_order.trade.__enter__.assert_called_with()
        mock_order.trade.__exit__.assert_called_with(None, None, None)

    @mock.patch("flumine.execution.betfairexecution.BetfairExecution.cancel")
    @mock.patch("flumine.execution.betfairexecution.BetfairExecution._execution_helper")
    def test_execute_cancel_error(self, mock__execution_helper, mock_place):
        mock_session = mock.Mock()
        mock_order_package = mock.Mock()
        mock_order_package.info = {}
        mock__execution_helper.return_value = None
        self.execution.execute_cancel(mock_order_package, mock_session)
        mock__execution_helper.assert_called_with(
            mock_place, mock_order_package, mock_session
        )

    def test_cancel(self):
        mock_session = mock.Mock()
        mock_order_package = mock.Mock()
        mock_order_package.cancel_instructions = [1, 2]
        self.assertEqual(
            self.execution.cancel(mock_order_package, mock_session),
            mock_order_package.client.betting_client.betting.cancel_orders(),
        )

    def test_cancel_empty(self):
        mock_session = mock.Mock()
        mock_order_package = mock.Mock()
        mock_order_package.info = {}
        mock_order_package.cancel_instructions = []
        with self.assertRaises(OrderExecutionError):
            self.execution.cancel(mock_order_package, mock_session)

    @mock.patch("flumine.execution.betfairexecution.BetfairExecution._order_logger")
    @mock.patch("flumine.execution.betfairexecution.BetfairExecution.update")
    @mock.patch("flumine.execution.betfairexecution.BetfairExecution._execution_helper")
    def test_execute_update_success(
        self, mock__execution_helper, mock_update, mock__order_logger
    ):
        mock_session = mock.Mock()
        mock_order = mock.Mock()
        mock_order.trade.__enter__ = mock.Mock()
        mock_order.trade.__exit__ = mock.Mock()
        mock_order.bet_id = 123
        mock_order_package = mock.Mock()
        mock_order_package.__iter__ = mock.Mock(return_value=iter([mock_order]))
        mock_order_package.info = {}
        mock_report = mock.Mock()
        mock_instruction_report = mock.Mock()
        mock_instruction_report.status = "SUCCESS"
        mock_report.update_instruction_reports = [mock_instruction_report]
        mock__execution_helper.return_value = mock_report
        self.execution.execute_update(mock_order_package, mock_session)
        mock__execution_helper.assert_called_with(
            mock_update, mock_order_package, mock_session
        )
        mock__order_logger.assert_called_with(
            mock_order, mock_instruction_report, OrderPackageType.UPDATE
        )
        mock_order.executable.assert_called_with()
        mock_order.trade.__enter__.assert_called_with()
        mock_order.trade.__exit__.assert_called_with(None, None, None)

    @mock.patch("flumine.execution.betfairexecution.BetfairExecution._order_logger")
    @mock.patch("flumine.execution.betfairexecution.BetfairExecution.update")
    @mock.patch("flumine.execution.betfairexecution.BetfairExecution._execution_helper")
    def test_execute_update_failure(
        self, mock__execution_helper, mock_update, mock__order_logger
    ):
        mock_session = mock.Mock()
        mock_order = mock.Mock()
        mock_order.trade.__enter__ = mock.Mock()
        mock_order.trade.__exit__ = mock.Mock()
        mock_order.bet_id = 123
        mock_order_package = mock.Mock()
        mock_order_package.__iter__ = mock.Mock(return_value=iter([mock_order]))
        mock_order_package.info = {}
        mock_report = mock.Mock()
        mock_instruction_report = mock.Mock()
        mock_instruction_report.status = "FAILURE"
        mock_report.update_instruction_reports = [mock_instruction_report]
        mock__execution_helper.return_value = mock_report
        self.execution.execute_update(mock_order_package, mock_session)
        mock__execution_helper.assert_called_with(
            mock_update, mock_order_package, mock_session
        )
        mock__order_logger.assert_called_with(
            mock_order, mock_instruction_report, OrderPackageType.UPDATE
        )
        mock_order.executable.assert_called_with()
        mock_order.trade.__enter__.assert_called_with()
        mock_order.trade.__exit__.assert_called_with(None, None, None)
        mock_order_package.client.add_transaction.assert_called_with(1, failed=True)

    @mock.patch("flumine.execution.betfairexecution.BetfairExecution._order_logger")
    @mock.patch("flumine.execution.betfairexecution.BetfairExecution.update")
    @mock.patch("flumine.execution.betfairexecution.BetfairExecution._execution_helper")
    def test_execute_update_timeout(
        self, mock__execution_helper, mock_update, mock__order_logger
    ):
        mock_session = mock.Mock()
        mock_order = mock.Mock()
        mock_order.trade.__enter__ = mock.Mock()
        mock_order.trade.__exit__ = mock.Mock()
        mock_order.bet_id = 123
        mock_order_package = mock.Mock()
        mock_order_package.__iter__ = mock.Mock(return_value=iter([mock_order]))
        mock_order_package.info = {}
        mock_report = mock.Mock()
        mock_instruction_report = mock.Mock()
        mock_instruction_report.status = "TIMEOUT"
        mock_report.update_instruction_reports = [mock_instruction_report]
        mock__execution_helper.return_value = mock_report
        self.execution.execute_update(mock_order_package, mock_session)
        mock__execution_helper.assert_called_with(
            mock_update, mock_order_package, mock_session
        )
        mock__order_logger.assert_called_with(
            mock_order, mock_instruction_report, OrderPackageType.UPDATE
        )
        mock_order.executable.assert_called_with()
        mock_order.trade.__enter__.assert_called_with()
        mock_order.trade.__exit__.assert_called_with(None, None, None)

    @mock.patch("flumine.execution.betfairexecution.BetfairExecution.update")
    @mock.patch("flumine.execution.betfairexecution.BetfairExecution._execution_helper")
    def test_execute_update_error(self, mock__execution_helper, mock_update):
        mock_session = mock.Mock()
        mock_order_package = mock.Mock()
        mock_order_package.info = {}
        mock__execution_helper.return_value = None
        self.execution.execute_update(mock_order_package, mock_session)
        mock__execution_helper.assert_called_with(
            mock_update, mock_order_package, mock_session
        )

    def test_update(self):
        mock_session = mock.Mock()
        mock_order_package = mock.Mock()
        self.assertEqual(
            self.execution.update(mock_order_package, mock_session),
            mock_order_package.client.betting_client.betting.update_orders(),
        )

    @mock.patch("flumine.execution.betfairexecution.BetfairExecution._order_logger")
    @mock.patch("flumine.execution.betfairexecution.BetfairExecution.replace")
    @mock.patch("flumine.execution.betfairexecution.BetfairExecution._execution_helper")
    def test_execute_replace_success(
        self, mock__execution_helper, mock_replace, mock__order_logger
    ):
        mock_market = mock.Mock()
        mock_market.transaction.__enter__ = mock.Mock()
        # mock_market.transaction.__exit__ = mock.Mock()
        self.mock_flumine.markets.markets = {"1.23": mock_market}
        mock_session = mock.Mock()
        mock_order = mock.Mock(market_id="1.23", bet_id=123)
        mock_order.trade.__enter__ = mock.Mock()
        mock_order.trade.__exit__ = mock.Mock()
        mock_order_package = mock.MagicMock(market_id="1.23", info={})
        mock_order_package.__len__.return_value = 1
        mock_order_package.__iter__ = mock.Mock(return_value=iter([mock_order]))
        mock_report = mock.Mock()
        mock_instruction_report = mock.Mock()
        mock_instruction_report.cancel_instruction_reports.status = "SUCCESS"
        mock_instruction_report.place_instruction_reports.status = "SUCCESS"
        mock_report.replace_instruction_reports = [mock_instruction_report]
        mock__execution_helper.return_value = mock_report
        self.execution.execute_replace(mock_order_package, mock_session)
        mock__execution_helper.assert_called_with(
            mock_replace, mock_order_package, mock_session
        )
        # mock__order_logger.assert_called_with(
        #     mock_order, mock_instruction_report, OrderPackageType.CANCEL
        # )
        mock_order.execution_complete.assert_called_with()
        replacement_order = mock_order.trade.create_order_replacement()
        replacement_order.executable.assert_called_with()
        mock_market.place_order.assert_called_with(
            replacement_order, execute=False, client=mock_order.client
        )
        mock__order_logger.assert_called_with(
            replacement_order,
            mock_instruction_report.place_instruction_reports,
            OrderPackageType.REPLACE,
        )
        mock_order.trade.__enter__.assert_called_with()
        mock_order.trade.__exit__.assert_called_with(None, None, None)
        mock_order_package.client.add_transaction.assert_called_with(1)

    @mock.patch("flumine.execution.betfairexecution.BetfairExecution.replace")
    @mock.patch("flumine.execution.betfairexecution.BetfairExecution._execution_helper")
    def test_execute_replace_error(self, mock__execution_helper, mock_replace):
        mock_session = mock.Mock()
        mock_order_package = mock.Mock()
        mock_order_package.info = {}
        mock__execution_helper.return_value = None
        self.execution.execute_replace(mock_order_package, mock_session)
        mock__execution_helper.assert_called_with(
            mock_replace, mock_order_package, mock_session
        )

    def test_replace(self):
        mock_session = mock.Mock()
        mock_order_package = mock.Mock()
        self.assertEqual(
            self.execution.replace(mock_order_package, mock_session),
            mock_order_package.client.betting_client.betting.replace_orders(),
        )

    @mock.patch(
        "flumine.execution.betfairexecution.BetfairExecution._return_http_session"
    )
    def test__execution_helper(self, mock__return_http_session):
        mock_trading_function = mock.Mock()
        mock_trading_function.__name__ = "test"
        mock_session = mock.Mock()
        mock_order_package = mock.Mock(
            elapsed_seconds=0.001, orders=[mock.Mock(elapsed_seconds_created=1)]
        )
        mock_order_package.info = {}
        self.execution._execution_helper(
            mock_trading_function, mock_order_package, mock_session
        )
        mock_trading_function.assert_called_with(mock_order_package, mock_session)
        mock__return_http_session.assert_called_with(mock_session)

    @mock.patch(
        "flumine.execution.betfairexecution.BetfairExecution._return_http_session"
    )
    def test__execution_helper_warning(self, mock__return_http_session):
        mock_trading_function = mock.Mock()
        mock_trading_function.__name__ = "test"
        mock_session = mock.Mock()
        mock_order_package = mock.Mock(
            elapsed_seconds=0.2,
            retry_count=0,
            orders=[mock.Mock(elapsed_seconds_created=1)],
        )
        mock_order_package.info = {}
        self.execution._execution_helper(
            mock_trading_function, mock_order_package, mock_session
        )
        mock_trading_function.assert_called_with(mock_order_package, mock_session)
        mock__return_http_session.assert_called_with(mock_session)

    @mock.patch(
        "flumine.execution.betfairexecution.BetfairExecution._return_http_session"
    )
    def test__execution_helper_empty(self, mock__return_http_session):
        mock_trading_function = mock.Mock()
        mock_trading_function.__name__ = "test"
        mock_session = mock.Mock()
        mock_order_package = mock.Mock(orders=[], elapsed_seconds=0.001)
        mock_order_package.info = {}
        self.execution._execution_helper(
            mock_trading_function, mock_order_package, mock_session
        )
        mock_trading_function.assert_not_called()
        mock__return_http_session.assert_called_with(mock_session)

    @mock.patch("flumine.execution.betfairexecution.BetfairExecution.handler")
    @mock.patch(
        "flumine.execution.betfairexecution.BetfairExecution._return_http_session"
    )
    def test__execution_helper_error(self, mock__return_http_session, mock_handler):
        mock_trading_function = mock.Mock()
        mock_trading_function.__name__ = "test"
        mock_trading_function.side_effect = BetfairError()
        mock_session = mock.Mock()
        mock_order_package = mock.Mock(
            elapsed_seconds=0.001, orders=[mock.Mock(elapsed_seconds_created=1)]
        )
        mock_order_package.info = {}
        mock_order_package.retry.return_value = True
        self.assertIsNone(
            self.execution._execution_helper(
                mock_trading_function, mock_order_package, mock_session
            )
        )
        mock_trading_function.assert_called_with(mock_order_package, mock_session)
        mock__return_http_session.assert_called_with(mock_session, err=True)
        mock_handler.assert_called_with(mock_order_package)

    @mock.patch("flumine.execution.betfairexecution.BetfairExecution.handler")
    @mock.patch(
        "flumine.execution.betfairexecution.BetfairExecution._return_http_session"
    )
    def test__execution_helper_error_no_retry(
        self, mock__return_http_session, mock_handler
    ):
        mock_trading_function = mock.Mock()
        mock_trading_function.__name__ = "test"
        mock_trading_function.side_effect = BetfairError()
        mock_session = mock.Mock()
        mock_order_package = mock.Mock(
            elapsed_seconds=0.001, orders=[mock.Mock(elapsed_seconds_created=1)]
        )
        mock_order_package.info = {}
        mock_order_package.retry.return_value = False
        self.assertIsNone(
            self.execution._execution_helper(
                mock_trading_function, mock_order_package, mock_session
            )
        )
        mock_trading_function.assert_called_with(mock_order_package, mock_session)
        mock__return_http_session.assert_called_with(mock_session, err=True)
        mock_handler.put.assert_not_called()
        mock_order_package.reset_orders.assert_called_with()

    @mock.patch("flumine.execution.betfairexecution.BetfairExecution.handler")
    @mock.patch(
        "flumine.execution.betfairexecution.BetfairExecution._return_http_session"
    )
    def test__execution_helper_error_no_retry_place(
        self, mock__return_http_session, mock_handler
    ):
        mock_trading_function = mock.Mock()
        mock_trading_function.__name__ = "test"
        mock_trading_function.side_effect = BetfairError()
        mock_session = mock.Mock()
        mock_order_package = mock.Mock(
            elapsed_seconds=0.001,
            package_type=OrderPackageType.PLACE,
            orders=[mock.Mock(elapsed_seconds_created=1)],
        )
        mock_order_package.info = {}
        mock_order_package.retry.return_value = False
        self.assertIsNone(
            self.execution._execution_helper(
                mock_trading_function, mock_order_package, mock_session
            )
        )
        mock_trading_function.assert_called_with(mock_order_package, mock_session)
        mock__return_http_session.assert_called_with(mock_session, err=True)
        mock_handler.put.assert_not_called()
        mock_order_package.reset_orders.assert_called_with(complete=True)

    @mock.patch(
        "flumine.execution.betfairexecution.BetfairExecution._return_http_session"
    )
    def test__execution_helper_unknown_error(self, mock__return_http_session):
        mock_trading_function = mock.Mock()
        mock_trading_function.__name__ = "test"
        mock_trading_function.side_effect = ValueError()
        mock_session = mock.Mock()
        mock_order_package = mock.Mock(
            elapsed_seconds=0.001, orders=[mock.Mock(elapsed_seconds_created=1)]
        )
        mock_order_package.info = {}
        mock_order_package.retry.return_value = True
        self.assertIsNone(
            self.execution._execution_helper(
                mock_trading_function, mock_order_package, mock_session
            )
        )
        mock_trading_function.assert_called_with(mock_order_package, mock_session)
        mock__return_http_session.assert_called_with(mock_session, err=True)


class SimulatedExecutionTest(unittest.TestCase):
    def setUp(self) -> None:
        self.mock_market = mock.Mock()
        self.mock_flumine = mock.Mock()
        self.mock_flumine.markets.markets = {"1.23": self.mock_market}
        self.execution = SimulatedExecution(self.mock_flumine)

    def test_init(self):
        self.assertEqual(self.execution.EXCHANGE, ExchangeType.SIMULATED)

    @mock.patch("flumine.execution.simulatedexecution.SimulatedExecution.execute_place")
    def test_handler_paper_trade(self, mock_execute_place):
        mock_thread_pool = mock.Mock()
        self.execution._thread_pool = mock_thread_pool
        mock_order_package = mock.Mock()
        mock_order_package.client.paper_trade = True
        mock_order_package.package_type = OrderPackageType.PLACE
        self.execution.handler(mock_order_package)
        mock_thread_pool.submit.assert_called_with(
            mock_execute_place, mock_order_package, None
        )

    @mock.patch("flumine.execution.simulatedexecution.SimulatedExecution.execute_place")
    def test_handler_place(self, mock_execute_place):
        mock_order_package = mock.Mock()
        mock_order_package.client.paper_trade = False
        mock_order_package.package_type = OrderPackageType.PLACE
        self.execution.handler(mock_order_package)
        mock_execute_place.assert_called_with(mock_order_package, http_session=None)

    @mock.patch(
        "flumine.execution.simulatedexecution.SimulatedExecution.execute_cancel"
    )
    def test_handler_cancel(self, mock_execute_cancel):
        mock_order_package = mock.Mock()
        mock_order_package.client.paper_trade = False
        mock_order_package.package_type = OrderPackageType.PLACE.CANCEL
        self.execution.handler(mock_order_package)
        mock_execute_cancel.assert_called_with(mock_order_package, http_session=None)

    @mock.patch(
        "flumine.execution.simulatedexecution.SimulatedExecution.execute_update"
    )
    def test_handler_update(self, mock_execute_update):
        mock_order_package = mock.Mock()
        mock_order_package.client.paper_trade = False
        mock_order_package.package_type = OrderPackageType.UPDATE
        self.execution.handler(mock_order_package)
        mock_execute_update.assert_called_with(mock_order_package, http_session=None)

    @mock.patch(
        "flumine.execution.simulatedexecution.SimulatedExecution.execute_replace"
    )
    def test_handler_replace(self, mock_execute_replace):
        mock_order_package = mock.Mock()
        mock_order_package.client.paper_trade = False
        mock_order_package.package_type = OrderPackageType.REPLACE
        self.execution.handler(mock_order_package)
        mock_execute_replace.assert_called_with(mock_order_package, http_session=None)

    @mock.patch("flumine.execution.simulatedexecution.SimulatedExecution._order_logger")
    def test_execute_place_success(self, mock__order_logger):
        mock_order = mock.Mock()
        mock_order.trade.__enter__ = mock.Mock()
        mock_order.trade.__exit__ = mock.Mock()
        mock_order_package = mock.MagicMock(market_id="1.23")
        mock_order_package.__len__.return_value = 1
        mock_order_package.client.paper_trade = False
        mock_order_package.__iter__ = mock.Mock(return_value=iter([mock_order]))
        mock_order_package.place_instructions = [1]
        mock_order_package.info = {}
        mock_sim_resp = mock.Mock()
        mock_sim_resp.status = "SUCCESS"
        mock_order.simulated.place.return_value = mock_sim_resp
        self.execution.execute_place(mock_order_package, None)
        mock_order.simulated.place.assert_called_with(
            mock_order_package,
            self.mock_market.market_book,
            1,
            self.execution._bet_id,
        )
        mock__order_logger.assert_called_with(
            mock_order, mock_sim_resp, mock_order_package.package_type
        )
        mock_order.executable.assert_called_with()
        mock_order.trade.__enter__.assert_called_with()
        mock_order.trade.__exit__.assert_called_with(None, None, None)
        mock_order_package.client.add_transaction.assert_called_with(1)

    @mock.patch("flumine.execution.simulatedexecution.SimulatedExecution._order_logger")
    def test_execute_place_failure(self, mock__order_logger):
        mock_order = mock.Mock()
        mock_order.trade.__enter__ = mock.Mock()
        mock_order.trade.__exit__ = mock.Mock()
        mock_order_package = mock.MagicMock(market_id="1.23")
        mock_order_package.__len__.return_value = 1
        mock_order_package.client.paper_trade = False
        mock_order_package.__iter__ = mock.Mock(return_value=iter([mock_order]))
        mock_order_package.place_instructions = [1]
        mock_order_package.info = {}
        mock_sim_resp = mock.Mock()
        mock_sim_resp.status = "FAILURE"
        mock_order.simulated.place.return_value = mock_sim_resp
        self.execution.execute_place(mock_order_package, None)
        mock_order.simulated.place.assert_called_with(
            mock_order_package,
            self.mock_market.market_book,
            1,
            self.execution._bet_id,
        )
        mock__order_logger.assert_called_with(
            mock_order, mock_sim_resp, mock_order_package.package_type
        )
        mock_order.execution_complete.assert_called_with()
        mock_order.trade.__enter__.assert_called_with()
        mock_order.trade.__exit__.assert_called_with(None, None, None)
        mock_order_package.client.add_transaction.assert_called_with(1)

    @mock.patch("flumine.execution.simulatedexecution.time")
    def test_execute_place_paper_trade(self, mock_time):
        mock_order_package = mock.MagicMock(
            market_id="1.23", place_instructions=[], bet_delay=1
        )
        mock_order_package.__len__.return_value = 1
        mock_order_package.__iter__ = mock.Mock(return_value=iter([]))
        mock_order_package.client.paper_trade = True
        self.execution.execute_place(mock_order_package, None)
        mock_time.sleep.assert_called_with(config.place_latency + 1)
        mock_order_package.client.add_transaction.assert_called_with(1)

    @mock.patch("flumine.execution.simulatedexecution.SimulatedExecution._order_logger")
    def test_execute_cancel(self, mock__order_logger):
        mock_client = mock.Mock()
        mock_order = mock.Mock(client=mock_client, size_cancelled=2, size_remaining=0)
        mock_order.trade.__enter__ = mock.Mock()
        mock_order.trade.__exit__ = mock.Mock()
        mock_order_package = mock.MagicMock(market_id="1.23", client=mock_client)
        mock_order_package.__len__.return_value = 1
        mock_order_package.client.paper_trade = False
        mock_order_package.__iter__ = mock.Mock(return_value=iter([mock_order]))
        mock_sim_resp = mock.Mock()
        mock_sim_resp.status = "SUCCESS"
        mock_order.simulated.cancel.return_value = mock_sim_resp
        self.execution.execute_cancel(mock_order_package, None)
        mock_order.simulated.cancel.assert_called_with(self.mock_market.market_book)
        mock__order_logger.assert_called_with(
            mock_order, mock_sim_resp, mock_order_package.package_type
        )
        mock_order.execution_complete.assert_called_with()
        mock_order.trade.__enter__.assert_called_with()
        mock_order.trade.__exit__.assert_called_with(None, None, None)

    @mock.patch("flumine.execution.simulatedexecution.SimulatedExecution._order_logger")
    def test_execute_cancel_executable(self, mock__order_logger):
        mock_client = mock.Mock()
        mock_order = mock.Mock(client=mock_client, size_cancelled=2, size_remaining=2)
        mock_order.trade.__enter__ = mock.Mock()
        mock_order.trade.__exit__ = mock.Mock()
        mock_order_package = mock.MagicMock(market_id="1.23", client=mock_client)
        mock_order_package.__len__.return_value = 1
        mock_order_package.client.paper_trade = False
        mock_order_package.__iter__ = mock.Mock(return_value=iter([mock_order]))
        mock_sim_resp = mock.Mock()
        mock_sim_resp.status = "SUCCESS"
        mock_order.simulated.cancel.return_value = mock_sim_resp
        self.execution.execute_cancel(mock_order_package, None)
        mock_order.simulated.cancel.assert_called_with(self.mock_market.market_book)
        mock__order_logger.assert_called_with(
            mock_order, mock_sim_resp, mock_order_package.package_type
        )
        mock_order.executable.assert_called_with()
        mock_order.trade.__enter__.assert_called_with()
        mock_order.trade.__exit__.assert_called_with(None, None, None)

    @mock.patch("flumine.execution.simulatedexecution.SimulatedExecution._order_logger")
    def test_execute_cancel_failure(self, mock__order_logger):
        mock_client = mock.Mock()
        mock_order = mock.Mock(client=mock_client)
        mock_order.trade.__enter__ = mock.Mock()
        mock_order.trade.__exit__ = mock.Mock()
        mock_order_package = mock.MagicMock(market_id="1.23", client=mock_client)
        mock_order_package.__len__.return_value = 1
        mock_order_package.client.paper_trade = False
        mock_order_package.__iter__ = mock.Mock(return_value=iter([mock_order]))
        mock_sim_resp = mock.Mock()
        mock_sim_resp.status = "FAILURE"
        mock_order.simulated.cancel.return_value = mock_sim_resp
        self.execution.execute_cancel(mock_order_package, None)
        mock_order.simulated.cancel.assert_called_with(self.mock_market.market_book)
        mock__order_logger.assert_called_with(
            mock_order, mock_sim_resp, mock_order_package.package_type
        )
        mock_order.executable.assert_called_with()
        mock_order.trade.__enter__.assert_called_with()
        mock_order.trade.__exit__.assert_called_with(None, None, None)
        mock_order_package.client.add_transaction.assert_called_with(1, failed=True)

    @mock.patch("flumine.execution.simulatedexecution.time")
    def test_execute_cancel_paper_trade(self, mock_time):
        mock_order_package = mock.Mock(market_id="1.23", bet_delay=1)
        mock_order_package.__iter__ = mock.Mock(return_value=iter([]))
        mock_order_package.client.paper_trade = True
        self.execution.execute_cancel(mock_order_package, None)
        mock_time.sleep.assert_called_with(config.cancel_latency)

    @mock.patch("flumine.execution.simulatedexecution.SimulatedExecution._order_logger")
    def test_execute_update(self, mock__order_logger):
        mock_client = mock.Mock()
        mock_order = mock.Mock(client=mock_client)
        mock_order.trade.__enter__ = mock.Mock()
        mock_order.trade.__exit__ = mock.Mock()
        mock_order_package = mock.Mock(market_id="1.23", client=mock_client)
        mock_order_package.client.paper_trade = False
        mock_order_package.__iter__ = mock.Mock(return_value=iter([mock_order]))
        mock_order_package.update_instructions = ["PERSIST"]
        mock_order_package.info = {}
        mock_sim_resp = mock.Mock()
        mock_sim_resp.status = "SUCCESS"
        mock_order.simulated.update.return_value = mock_sim_resp
        self.execution.execute_update(mock_order_package, None)
        mock_order.simulated.update.assert_called_with(
            self.mock_market.market_book, "PERSIST"
        )
        mock__order_logger.assert_called_with(
            mock_order, mock_sim_resp, mock_order_package.package_type
        )
        mock_order.executable.assert_called_with()
        mock_order.trade.__enter__.assert_called_with()
        mock_order.trade.__exit__.assert_called_with(None, None, None)

    @mock.patch("flumine.execution.simulatedexecution.SimulatedExecution._order_logger")
    def test_execute_update_failure(self, mock__order_logger):
        mock_client = mock.Mock()
        mock_order = mock.Mock(client=mock_client)
        mock_order.trade.__enter__ = mock.Mock()
        mock_order.trade.__exit__ = mock.Mock()
        mock_order_package = mock.Mock(market_id="1.23", client=mock_client)
        mock_order_package.client.paper_trade = False
        mock_order_package.__iter__ = mock.Mock(return_value=iter([mock_order]))
        mock_order_package.update_instructions = ["PERSIST"]
        mock_order_package.info = {}
        mock_sim_resp = mock.Mock()
        mock_sim_resp.status = "FAILURE"
        mock_order.simulated.update.return_value = mock_sim_resp
        self.execution.execute_update(mock_order_package, None)
        mock_order.simulated.update.assert_called_with(
            self.mock_market.market_book, "PERSIST"
        )
        mock__order_logger.assert_called_with(
            mock_order, mock_sim_resp, mock_order_package.package_type
        )
        mock_order.executable.assert_called_with()
        mock_order.trade.__enter__.assert_called_with()
        mock_order.trade.__exit__.assert_called_with(None, None, None)
        mock_order_package.client.add_transaction.assert_called_with(1, failed=True)

    @mock.patch("flumine.execution.simulatedexecution.time")
    def test_execute_update_paper_trade(self, mock_time):
        mock_order_package = mock.Mock(
            market_id="1.23", update_instructions=[], bet_delay=1
        )
        mock_order_package.__iter__ = mock.Mock(return_value=iter([]))
        mock_order_package.client.paper_trade = True
        self.execution.execute_update(mock_order_package, None)
        mock_time.sleep.assert_called_with(config.update_latency)

    @mock.patch("flumine.execution.simulatedexecution.SimulatedExecution._order_logger")
    def test_execute_replace(self, mock__order_logger):
        mock_client = mock.Mock()
        mock_order = mock.Mock(client=mock_client)
        mock_order.trade.__enter__ = mock.Mock()
        mock_order.trade.__exit__ = mock.Mock()
        mock_order.market_id = "1.234"
        mock_order_package = mock.MagicMock(market_id="1.23", client=mock_client)
        mock_order_package.__len__.return_value = 1
        mock_order_package.client.paper_trade = False
        mock_order_package.__iter__ = mock.Mock(return_value=iter([mock_order]))
        mock_order_package.replace_instructions = [{"newPrice": 2.03}]
        mock_replacement_order = mock.Mock()
        mock_replacement_order_package = mock.Mock()
        mock_replacement_order_package.__iter__ = mock.Mock(
            return_value=iter([mock_order])
        )
        mock_sim_resp = mock.Mock()
        mock_sim_resp.status = "SUCCESS"
        mock_order.simulated.cancel.return_value = mock_sim_resp
        mock_replacement_order.simulated.place.return_value = mock_sim_resp
        mock_order.trade.create_order_replacement.return_value = mock_replacement_order
        self.execution.execute_replace(mock_order_package, None)
        mock_order.simulated.cancel.assert_called_with(self.mock_market.market_book)
        mock_replacement_order.simulated.place.assert_called_with(
            mock_order_package,
            self.mock_market.market_book,
            {"newPrice": 2.03},
            self.execution._bet_id,
        )
        mock__order_logger.assert_has_calls(
            [
                call(mock_order, mock_sim_resp, OrderPackageType.CANCEL),
                call(
                    mock_replacement_order,
                    mock_sim_resp,
                    mock_order_package.package_type,
                ),
            ]
        )
        mock_order.execution_complete.assert_called_with()
        # mock_replacement_order.executable.assert_called_with()
        mock_order.trade.__enter__.assert_called_with()
        mock_order.trade.__exit__.assert_called_with(None, None, None)
        mock_order_package.client.add_transaction.assert_called_with(1)

    @mock.patch("flumine.execution.simulatedexecution.SimulatedExecution._order_logger")
    def test_execute_replace_failure(self, mock__order_logger):
        mock_client = mock.Mock()
        mock_order = mock.Mock(client=mock_client)
        mock_order.trade.__enter__ = mock.Mock()
        mock_order.trade.__exit__ = mock.Mock()
        mock_order_package = mock.MagicMock(market_id="1.23", client=mock_client)
        mock_order_package.__len__.return_value = 1
        mock_order_package.client.paper_trade = False
        mock_order_package.__iter__ = mock.Mock(return_value=iter([mock_order]))
        mock_order_package.replace_instructions = [{"newPrice": 2.54}]
        mock_replacement_order = mock.Mock()
        mock_replacement_order_package = mock.Mock()
        mock_replacement_order_package.__iter__ = mock.Mock(
            return_value=iter([mock_order])
        )
        mock_sim_resp = mock.Mock()
        mock_sim_resp.status = "FAILURE"
        mock_order.simulated.cancel.return_value = mock_sim_resp
        mock_replacement_order.simulated.place.return_value = mock_sim_resp
        mock_order.trade.create_order_replacement.return_value = mock_replacement_order
        self.execution.execute_replace(mock_order_package, None)
        mock_order.simulated.cancel.assert_called_with(self.mock_market.market_book)
        mock_replacement_order.simulated.place.assert_not_called()
        mock__order_logger.assert_called_with(
            mock_order, mock_sim_resp, OrderPackageType.CANCEL
        )
        mock_order.executable.assert_called_with()
        mock_replacement_order.executable.assert_not_called()
        mock_order.trade.__enter__.assert_called_with()
        mock_order.trade.__exit__.assert_called_with(None, None, None)

    @mock.patch("flumine.execution.simulatedexecution.time")
    def test_execute_replace_paper_trade(self, mock_time):
        mock_order_package = mock.MagicMock(
            market_id="1.23", replace_instructions=[], bet_delay=1
        )
        mock_order_package.__len__.return_value = 1
        mock_order_package.__iter__ = mock.Mock(return_value=iter([]))
        mock_order_package.client.paper_trade = True
        self.execution.execute_replace(mock_order_package, None)
        mock_time.sleep.assert_called_with(config.replace_latency + 1)



from typing import Optional
from betfairlightweight.metadata import currency_parameters
from betfairlightweight.resources.accountresources import AccountDetails

from .baseclient import BaseClient
from .clients import ExchangeType


class SimulatedClient(BaseClient):
    """
    Simulated betting client.
    """

    EXCHANGE = ExchangeType.SIMULATED
    DISCOUNT_RATE = 0
    CURRENCY_CODE = "GBP"

    def login(self) -> None:
        return

    def keep_alive(self) -> None:
        return

    def logout(self) -> None:
        return

    def update_account_details(self) -> None:
        self.account_details = AccountDetails(
            **{"discountRate": self.DISCOUNT_RATE, "currencyCode": self.CURRENCY_CODE}
        )

    @property
    def min_bet_size(self) -> Optional[float]:
        if self.account_details:
            return currency_parameters[self.account_details.currency_code][
                "min_bet_size"
            ]

    @property
    def min_bsp_liability(self) -> Optional[float]:
        if self.account_details:
            return currency_parameters[self.account_details.currency_code][
                "min_bsp_liability"
            ]

    @property
    def min_bet_payout(self) -> Optional[float]:
        if self.account_details:
            return currency_parameters[self.account_details.currency_code][
                "min_bet_payout"
            ]



import logging
import datetime
from typing import List, Optional
from betfairlightweight.resources.bettingresources import MarketBook, RunnerBook

from .utils import (
    SimulatedPlaceResponse,
    SimulatedCancelResponse,
    SimulatedUpdateResponse,
)
from ..utils import get_price, wap
from ..order.ordertype import OrderTypes
from .. import config

logger = logging.getLogger(__name__)


class SimulatedOrder:
    """
    Class to hold `simulated` order
    matching and status.
    """

    def __init__(self, order):
        self.order = order
        self.size_matched = 0
        self.average_price_matched = 0
        self.matched = []  # [[publishTime, price, size]..]
        self.size_cancelled = 0.0
        self.size_lapsed = 0.0
        self.size_voided = 0.0
        self.market_version = None  # version at place so we can lapse if needed
        self._piq = 0.0
        self._bsp_reconciled = False

    def __call__(self, market_book: MarketBook, traded: dict) -> None:
        # simulates order matching
        if self._bsp_reconciled is False and market_book.bsp_reconciled:
            if self.take_sp:
                runner = self._get_runner(market_book)
                self._process_sp(market_book.publish_time_epoch, runner)
                return
            else:
                self._bsp_reconciled = True

        if self.order.order_type.ORDER_TYPE == OrderTypes.LIMIT:
            if market_book.version != self.market_version:
                self.market_version = market_book.version  # update for next time
                if market_book.status == "SUSPENDED":  # Material change
                    if self.order.order_type.persistence_type == "LAPSE":
                        self.size_lapsed += self.size_remaining
                        return

            # todo estimated piq cancellations
            if traded:
                self._process_traded(market_book.publish_time_epoch, traded)

    def place(
        self, order_package, market_book: MarketBook, instruction: dict, bet_id: int
    ) -> SimulatedPlaceResponse:
        # simulates placeOrder request->matching->response
        # todo instruction/fillkill/timeInForce etc
        # validate market status
        if market_book.status != "OPEN":
            self.size_voided += self.size_remaining
            return self._create_place_response(
                None,
                status="FAILURE",
                error_code="ERROR_IN_ORDER",
            )

        # validate market version
        self.market_version = market_book.version
        if (
            order_package.market_version
            and order_package.market_version["version"] != self.market_version
        ):
            self.size_voided += self.size_remaining
            return self._create_place_response(
                None,
                status="FAILURE",
                error_code="ERROR_IN_ORDER",
            )

        runner = self._get_runner(market_book)
        # validate runner status
        if runner.status == "REMOVED":
            self.size_voided += self.size_remaining
            return self._create_place_response(
                None,
                status="FAILURE",
                error_code="RUNNER_REMOVED",
            )
        if self.order.order_type.ORDER_TYPE == OrderTypes.LIMIT:
            price = self.order.order_type.price
            size = self.order.order_type.size
            if size is None:
                raise NotImplementedError(
                    "Simulated betTargetSize placement not implemented"
                )
            if self.order.side == "BACK":
                available_to_back = get_price(runner.ex.available_to_back, 0) or 1.01
                if (
                    not order_package.client.best_price_execution
                    and available_to_back > price
                ):
                    self.size_lapsed += self.size_remaining
                    return self._create_place_response(
                        None,
                        status="FAILURE",
                        error_code="BET_LAPSED_PRICE_IMPROVEMENT_TOO_LARGE",
                    )
                elif available_to_back >= price:
                    self._process_price_matched(
                        market_book.publish_time_epoch,
                        price,
                        size,
                        runner.ex.available_to_back,
                    )
                    return self._create_place_response(bet_id)
                available = runner.ex.available_to_lay
            else:
                available_to_lay = get_price(runner.ex.available_to_lay, 0) or 1000
                if (
                    not order_package.client.best_price_execution
                    and available_to_lay < price
                ):
                    self.size_lapsed += self.size_remaining
                    return self._create_place_response(
                        None,
                        status="FAILURE",
                        error_code="BET_LAPSED_PRICE_IMPROVEMENT_TOO_LARGE",
                    )
                elif available_to_lay <= price:
                    self._process_price_matched(
                        market_book.publish_time_epoch,
                        price,
                        size,
                        runner.ex.available_to_lay,
                    )
                    return self._create_place_response(bet_id)
                available = runner.ex.available_to_back

            # calculate position in queue
            for avail in available:
                if avail["price"] == price:
                    self._piq = avail["size"]
                    break

            logger.debug(
                "Simulated order {0} PIQ: {1}".format(self.order.id, self._piq)
            )
            return self._create_place_response(bet_id)
        else:
            # validate BSP available, not reconciled and market not inplay
            if (
                market_book.market_definition.bsp_market is False
                or market_book.bsp_reconciled is True
                or market_book.inplay is True
            ):
                self.size_voided += self.size_remaining
                return self._create_place_response(
                    None,
                    status="FAILURE",
                    error_code="MARKET_NOT_OPEN_FOR_BSP_BETTING",
                )
            return self._create_place_response(bet_id)

    def _create_place_response(
        self,
        bet_id: Optional[int],
        status: str = "SUCCESS",
        order_status: str = None,
        error_code: str = None,
    ) -> SimulatedPlaceResponse:
        if self.order.client.simulated_full_match:
            if status == "SUCCESS" and self.size_remaining:
                self.matched.append(
                    [0, self.order.order_type.price, self.size_remaining]
                )
                self.size_matched, self.average_price_matched = wap(self.matched)
        if order_status is None:
            if self.size_remaining == 0:
                order_status = "EXECUTION_COMPLETE"
            else:
                order_status = "EXECUTABLE"
        return SimulatedPlaceResponse(
            status=status,
            order_status=order_status,
            bet_id=str(bet_id) if bet_id else bet_id,
            average_price_matched=self.average_price_matched,
            size_matched=self.size_matched,
            placed_date=datetime.datetime.utcnow(),
            error_code=error_code,
        )

    def cancel(self, market_book: MarketBook) -> SimulatedCancelResponse:
        # simulates cancelOrder request->cancel->response
        if market_book.status != "OPEN":
            return SimulatedCancelResponse(
                status="FAILURE",
                error_code="ERROR_IN_ORDER",
            )
        if self.order.order_type.ORDER_TYPE == OrderTypes.LIMIT:
            _size_reduction = (
                self.order.update_data.get("size_reduction") or self.size_remaining
            )
            _size_cancelled = min(
                _size_reduction, self.size_remaining
            )  # cancelled cannot be more than remaining (does not error)
            self.size_cancelled += _size_cancelled
            return SimulatedCancelResponse(
                status="SUCCESS",  # todo handle errors
                size_cancelled=_size_cancelled,
                cancelled_date=datetime.datetime.utcnow(),
            )
        else:
            return SimulatedCancelResponse(
                status="FAILURE",
                error_code="BET_ACTION_ERROR",
            )

    def update(
        self, market_book: MarketBook, instruction: dict
    ) -> SimulatedUpdateResponse:
        # simulates updateOrder request->update->response
        if market_book.status != "OPEN":
            return SimulatedUpdateResponse(
                status="FAILURE",
                error_code="ERROR_IN_ORDER",
            )
        # validate BSP available
        if market_book.market_definition.persistence_enabled is False:
            return SimulatedUpdateResponse(
                status="FAILURE",
                error_code="INVALID_PERSISTENCE_TYPE",
            )
        if (
            self.order.order_type.ORDER_TYPE == OrderTypes.LIMIT
            and self.size_remaining > 0
        ):
            self.order.order_type.persistence_type = instruction.get(
                "newPersistenceType"
            )
            return SimulatedUpdateResponse(status="SUCCESS")
        else:
            return SimulatedUpdateResponse(
                status="FAILURE",
                error_code="BET_ACTION_ERROR",
            )

    def _get_runner(self, market_book: MarketBook) -> RunnerBook:
        runner_dict = {
            (runner.selection_id, runner.handicap): runner
            for runner in market_book.runners
        }
        return runner_dict.get((self.order.selection_id, self.order.handicap))

    def _process_price_matched(
        self, publish_time: int, price: float, size: float, available: list
    ) -> None:
        # calculate matched on execution
        size_remaining = size
        for avail in available:
            if size_remaining == 0:
                break
            elif (self.side == "BACK" and price <= avail["price"]) or (
                self.side == "LAY" and price >= avail["price"]
            ):
                _size_remaining = size_remaining
                size_remaining = max(size_remaining - avail["size"], 0)
                if size_remaining == 0:
                    _size_matched = _size_remaining
                else:
                    _size_matched = avail["size"]
                _matched = [publish_time, avail["price"], round(_size_matched, 2)]
                self._update_matched(_matched)
            else:
                break

    def _process_sp(self, publish_time: int, runner: RunnerBook) -> None:
        # calculate matched on BSP reconciliation
        actual_sp = runner.sp.actual_sp
        if actual_sp:
            self._bsp_reconciled = True
            _order_type = self.order.order_type
            if _order_type.ORDER_TYPE == OrderTypes.LIMIT:
                if self.side == "BACK":
                    size = self.size_remaining
                else:
                    remaining_risk = (_order_type.price - 1.0) * self.size_remaining
                    client = self.order.client
                    if remaining_risk >= client.min_bsp_liability:
                        size = round(remaining_risk / (actual_sp - 1.0), 2)
                        # Cancel remaining size
                        self.size_cancelled += round(self.size_remaining - size, 2)
                    else:  # lapse
                        self.size_lapsed += self.size_remaining
                        self.order.execution_complete()
                        return
            elif _order_type.ORDER_TYPE == OrderTypes.LIMIT_ON_CLOSE:
                if self.side == "BACK":
                    if actual_sp < _order_type.price:
                        self.order.execution_complete()
                        return
                    size = _order_type.liability
                else:
                    if actual_sp > _order_type.price:
                        self.order.execution_complete()
                        return
                    size = round(
                        _order_type.liability / (actual_sp - 1), 2
                    )  # todo round 2dp down
            elif _order_type.ORDER_TYPE == OrderTypes.MARKET_ON_CLOSE:
                if self.side == "BACK":
                    size = _order_type.liability
                else:
                    size = round(_order_type.liability / (actual_sp - 1), 2)
            else:
                raise NotImplementedError()
            self._update_matched([publish_time, actual_sp, size])
            self.order.execution_complete()
        else:
            logger.warning(
                "SP not available for order {0} on runner {1}".format(
                    self.order.id, runner.selection_id
                )
            )

    def _process_traded(self, publish_time: int, traded: dict) -> None:
        # calculate matched on MarketBook update
        price = self.order.order_type.price
        side = self.side
        for traded_price, traded_size in traded.items():
            if logger.isEnabledFor(logging.DEBUG):
                logger.debug(
                    "Simulated order {0} traded: {1} - {2}".format(
                        self.order.id, traded_price, traded_size
                    )
                )
            if side == "BACK" and traded_price >= price:
                matched = self._calculate_process_traded(publish_time, traded_size)
                if matched:
                    traded[traded_price] = traded_size - matched
            elif side == "LAY" and traded_price <= price:
                matched = self._calculate_process_traded(publish_time, traded_size)
                if matched:
                    traded[traded_price] = traded_size - matched

    def _calculate_process_traded(self, publish_time: int, traded_size: float) -> float:
        _traded_size = traded_size / 2
        if self._piq - _traded_size < 0:
            size = _traded_size - self._piq
            size = round(min(self.size_remaining, size), 2)
            if size:
                self._update_matched(
                    [
                        publish_time,
                        self.order.order_type.price,
                        size,
                    ]
                )
            _matched = (self._piq + size) * 2
            self._piq = 0
            return _matched
        else:
            self._piq -= _traded_size
            if logger.isEnabledFor(logging.DEBUG):
                logger.debug(
                    "Simulated order {0} PIQ: {1}".format(self.order.id, self._piq)
                )
            return traded_size

    @property
    def take_sp(self) -> bool:
        if self.order.order_type.ORDER_TYPE == OrderTypes.LIMIT:
            if self.order.order_type.persistence_type == "MARKET_ON_CLOSE":
                return True
            return False
        else:
            return True

    @property
    def side(self) -> str:
        return self.order.side

    def _update_matched(self, data: List) -> None:
        logger.debug("Simulated order {0} matched: {1}".format(self.order.id, data))
        self.matched.append(data)
        self.size_matched, self.average_price_matched = wap(self.matched)

    @property
    def size_remaining(self) -> float:
        if self.order.order_type.ORDER_TYPE == OrderTypes.LIMIT:
            size = self.order.order_type.size or self.order.order_type.bet_target_size
            return round(
                size
                - self.size_matched
                - self.size_cancelled
                - self.size_lapsed
                - self.size_voided,
                2,
            )
        else:
            return 0.0

    @property
    def profit(self) -> float:
        number_of_dead_heat_winners = self.order.number_of_dead_heat_winners or 1
        if self.order.market_type == "EACH_WAY":
            divisor = self.order.each_way_divisor
            win = self.size_matched * (self.average_price_matched - 1)
            place = self.size_matched * (
                (self.average_price_matched - 1) * (1 / divisor)
            )
            if number_of_dead_heat_winners > 1:  # todo dead heat
                logger.error("Each Way Dead Heat Detected but not handled")
            if self.order.runner_status == "WINNER":
                profit = round(win + place, 2)
                if self.side == "BACK":
                    return profit
                else:
                    return -profit
            elif self.order.runner_status == "PLACED":
                if self.side == "BACK":
                    profit = place - self.size_matched
                    return round(profit, 2)
                else:
                    profit = self.size_matched - place
                    return round(profit, 2)
            elif self.order.runner_status == "LOSER":
                matched = round(self.size_matched * 2, 2)
                if self.side == "BACK":
                    return -matched
                else:
                    return matched
            else:
                return 0.0
        else:
            if self.order.runner_status == "WINNER":
                profit = (self.size_matched / number_of_dead_heat_winners) * (
                    self.average_price_matched - 1
                )
                if number_of_dead_heat_winners == 2:
                    profit = profit - (self.size_matched / number_of_dead_heat_winners)
                elif number_of_dead_heat_winners > 2:
                    profit = profit - (
                        self.size_matched
                        * (number_of_dead_heat_winners - 1)
                        / number_of_dead_heat_winners
                    )
                if self.side == "LAY":
                    profit = -profit
                return round(profit, 2)
            elif self.order.runner_status == "LOSER":
                if self.side == "BACK":
                    return -self.size_matched
                else:
                    return self.size_matched
            else:
                return 0.0

    @property
    def status(self) -> str:
        if self.take_sp:
            if self._bsp_reconciled:
                return "EXECUTION_COMPLETE"
            else:
                return "EXECUTABLE"
        else:
            if self.size_remaining:
                return "EXECUTABLE"
            else:
                return "EXECUTION_COMPLETE"

    @property
    def info(self) -> dict:
        return {
            "profit": self.profit,
            "piq": self._piq,
            "matched": self.matched,
        }

    def __bool__(self):
        return config.simulated or bool(
            self.order.client and self.order.client.paper_trade
        )



import unittest
from unittest import mock

from flumine.order.responses import Responses


class ResponsesTest(unittest.TestCase):
    def setUp(self):
        self.responses = Responses()

    def test_init(self):
        self.assertIsNotNone(self.responses.date_time_created)
        self.assertIsNone(self.responses.place_response)
        self.assertEqual(self.responses.cancel_responses, [])
        self.assertEqual(self.responses.replace_responses, [])
        self.assertEqual(self.responses.cancel_responses, [])
        self.assertIsNone(self.responses._date_time_placed)
        self.assertIsNone(self.responses.current_order)

    def test_placed(self):
        self.responses.placed(12)
        self.assertIsNotNone(self.responses._date_time_placed)
        self.assertEqual(self.responses.place_response, 12)

    def test_cancelled(self):
        self.responses.cancelled(12)
        self.assertEqual(self.responses.cancel_responses, [12])

    def test_replaced(self):
        self.responses.replaced(12)
        # self.assertIsNotNone(self.responses.date_time_placed)
        self.assertEqual(self.responses.replace_responses, [12])

    def test_updated(self):
        self.responses.updated(12)
        self.assertEqual(self.responses.update_responses, [12])

    def test_date_time_placed(self):
        self.assertIsNone(self.responses.date_time_placed)
        self.responses._date_time_placed = 1
        self.responses.current_order = mock.Mock(placed_date=2)
        self.assertEqual(self.responses.date_time_placed, 1)
        self.responses._date_time_placed = None
        self.assertEqual(self.responses.date_time_placed, 2)
        self.responses._date_time_placed = None
        self.responses.current_order = None
        self.assertIsNone(self.responses.date_time_placed)



import unittest
from unittest import mock

from flumine.markets.middleware import (
    Middleware,
    SimulatedMiddleware,
    RunnerAnalytics,
    OrderStatus,
    OrderTypes,
    WIN_MINIMUM_ADJUSTMENT_FACTOR,
    PLACE_MINIMUM_ADJUSTMENT_FACTOR,
    LIVE_STATUS,
    SimulatedSportsDataMiddleware,
)
from flumine.order.ordertype import MarketOnCloseOrder


class MiddlewareTest(unittest.TestCase):
    def setUp(self) -> None:
        self.middleware = Middleware()

    def test_call(self):
        self.middleware(None)

    def test_add_market(self):
        mock_market = mock.Mock()
        self.assertIsNone(self.middleware.add_market(mock_market))

    def test_remove_market(self):
        mock_market = mock.Mock()
        self.assertIsNone(self.middleware.remove_market(mock_market))


class SimulatedMiddlewareTest(unittest.TestCase):
    def setUp(self) -> None:
        self.middleware = SimulatedMiddleware()

    def test_init(self):
        self.assertEqual(self.middleware.markets, {})
        self.assertEqual(self.middleware._runner_removals, [])
        self.assertEqual(WIN_MINIMUM_ADJUSTMENT_FACTOR, 2.5)
        self.assertEqual(PLACE_MINIMUM_ADJUSTMENT_FACTOR, 0)
        self.assertEqual(
            LIVE_STATUS,
            [
                OrderStatus.EXECUTABLE,
                OrderStatus.CANCELLING,
                OrderStatus.UPDATING,
                OrderStatus.REPLACING,
            ],
        )

    @mock.patch(
        "flumine.markets.middleware.SimulatedMiddleware._process_simulated_orders"
    )
    @mock.patch("flumine.markets.middleware.SimulatedMiddleware._process_runner")
    def test_call(self, mock__process_runner, mock__process_simulated_orders):
        mock_market = mock.Mock(context={})
        mock_market_book = mock.Mock()
        mock_runner = mock.Mock(status="ACTIVE")
        mock_market_book.runners = [mock_runner]
        mock_market.market_book = mock_market_book
        self.middleware(mock_market)
        mock__process_runner.assert_called_with({}, mock_runner)
        self.assertEqual(mock_market.context, {"simulated": {}})
        mock__process_simulated_orders.assert_called_with(mock_market, {})

    @mock.patch(
        "flumine.markets.middleware.SimulatedMiddleware._process_simulated_orders"
    )
    @mock.patch(
        "flumine.markets.middleware.SimulatedMiddleware._process_runner_removal"
    )
    def test_call_non_runner(
        self, mock__process_runner_removal, mock__process_simulated_orders
    ):
        mock_market = mock.Mock(context={})
        mock_market_book = mock.Mock()
        mock_runner = mock.Mock(status="REMOVED")
        mock_market_book.runners = [mock_runner]
        mock_market.market_book = mock_market_book
        self.middleware(mock_market)
        self.assertEqual(
            self.middleware._runner_removals,
            [
                (
                    mock_runner.selection_id,
                    mock_runner.handicap,
                    mock_runner.adjustment_factor,
                )
            ],
        )
        mock__process_runner_removal.assert_called_with(
            mock_market,
            mock_runner.selection_id,
            mock_runner.handicap,
            mock_runner.adjustment_factor,
        )

    def test_remove_market(self):
        mock_market = mock.Mock(market_id="1.23")
        self.middleware.markets = {mock_market.market_id: []}
        self.middleware.remove_market(mock_market)
        self.middleware.remove_market(mock_market)
        self.assertEqual(self.middleware.markets, {})

    def test__process_runner_removal(self):
        mock_simulated = mock.MagicMock(matched=[[123, 8.6, 10]])
        mock_simulated.__bool__.return_value = True
        mock_order = mock.Mock(simulated=mock_simulated, info={})
        mock_simulated_two = mock.MagicMock(matched=[[123, 8.6, 10]])
        mock_simulated_two.__bool__.return_value = False
        mock_order_two = mock.Mock(simulated=mock_simulated_two, info={})
        mock_market = mock.Mock(blotter=[mock_order, mock_order_two])
        self.middleware._process_runner_removal(mock_market, 12345, 0, 16.2)
        self.assertEqual(mock_order.simulated.matched, [[123, 7.21, 10]])
        self.assertEqual(mock_order.simulated.average_price_matched, 7.21)
        self.assertEqual(mock_order_two.simulated.matched, [[123, 8.6, 10]])

    def test__process_runner_removal_under_limit(self):
        mock_simulated = mock.MagicMock(matched=[[123, 8.6, 10]])
        mock_simulated.__bool__.return_value = True
        mock_order = mock.Mock(simulated=mock_simulated)
        mock_market = mock.Mock(blotter=[mock_order])
        self.middleware._process_runner_removal(mock_market, 12345, 0, 2.4)
        self.assertEqual(mock_order.simulated.matched, [[123, 8.6, 10]])

    def test__process_runner_removal_void(self):
        mock_simulated = mock.MagicMock(matched=[[123, 8.6, 10]])
        mock_simulated.__bool__.return_value = True
        mock_order = mock.Mock(
            lookup=("1.23", 12345, 0), simulated=mock_simulated, info={}
        )
        mock_order.order_type.size = 10
        mock_order.order_type.ORDER_TYPE = OrderTypes.LIMIT
        mock_market = mock.Mock(market_id="1.23", blotter=[mock_order])
        self.middleware._process_runner_removal(mock_market, 12345, 0, 16.2)
        self.assertEqual(mock_order.simulated.size_matched, 0)
        self.assertEqual(mock_order.simulated.average_price_matched, 0)
        self.assertEqual(mock_order.simulated.matched, [])
        self.assertEqual(mock_order.simulated.size_voided, 10)

    def test__process_runner_removal_none(self):
        mock_simulated = mock.MagicMock(matched=[[123, 8.6, 10]])
        mock_simulated.__bool__.return_value = True
        mock_order = mock.Mock(simulated=mock_simulated)
        mock_market = mock.Mock(blotter=[mock_order])
        self.middleware._process_runner_removal(mock_market, 12345, 0, None)
        self.assertEqual(mock_order.simulated.matched, [[123, 8.6, 10]])

    def test__process_runner_removal_sp_win(self):
        order_type = MarketOnCloseOrder(liability=200)
        mock_order = mock.Mock(
            selection_id=1234,
            handicap=0,
            order_type=order_type,
            info={},
            side="LAY",
            current_order=mock.Mock(size_matched=0),
            average_price_matched=None,
        )

        mock_market_book = mock.Mock()
        mock_market_book.runners = [
            mock.Mock(selection_id=1234, handicap=0, adjustment_factor=20)
        ]
        mock_market = mock.Mock(
            market_type="WIN", blotter=[mock_order], market_book=mock_market_book
        )
        self.middleware._process_runner_removal(mock_market, 12345, 0, 50)

        # The liability of £200 is adjusted by the multiplier of 37.5%, which s
        # defined in the example here: https://github.com/betcode-org/flumine/issues/454
        self.assertEqual(mock_order.order_type.liability, 75)

    def test__process_runner_removal_sp_win_inplay(self):
        order_type = MarketOnCloseOrder(liability=200)
        mock_order = mock.Mock(
            selection_id=1234,
            handicap=0,
            order_type=order_type,
            info={},
            side="LAY",
            current_order=mock.Mock(size_matched=0),
            average_price_matched=10,
        )

        mock_market_book = mock.Mock()
        mock_market_book.runners = [
            mock.Mock(selection_id=1234, handicap=0, adjustment_factor=20)
        ]
        mock_market = mock.Mock(
            market_type="WIN", blotter=[mock_order], market_book=mock_market_book
        )
        self.middleware._process_runner_removal(mock_market, 12345, 0, 50)

        # The liability of £200 is adjusted by the multiplier of 37.5%, which s
        # defined in the example here: https://github.com/betcode-org/flumine/issues/454
        self.assertEqual(mock_order.order_type.liability, 75)
        # Size matched should be 75 / (10.0-1.0) \approx 8.33
        self.assertEqual(8.33, mock_order.current_order.size_matched)

    def test__process_runner_removal_sp_place(self):
        order_type = MarketOnCloseOrder(liability=200)
        mock_order = mock.Mock(
            selection_id=1234,
            handicap=0,
            order_type=order_type,
            info={},
            side="LAY",
            current_order=mock.Mock(size_matched=0),
            average_price_matched=None,
        )

        mock_market_book = mock.Mock()
        mock_market_book.runners = [
            mock.Mock(selection_id=1234, handicap=0, adjustment_factor=20)
        ]
        mock_market = mock.Mock(
            market_type="PLACE", blotter=[mock_order], market_book=mock_market_book
        )
        self.middleware._process_runner_removal(mock_market, 12345, 0, 50)

        # The liability of £200 is reduced by the non runner's adjustment factor of 50%
        self.assertEqual(mock_order.order_type.liability, 100)

    def test__process_runner_removal_sp_place_inplay(self):
        order_type = MarketOnCloseOrder(liability=200)
        mock_order = mock.Mock(
            selection_id=1234,
            handicap=0,
            order_type=order_type,
            info={},
            side="LAY",
            current_order=mock.Mock(size_matched=0),
            average_price_matched=10.0,
        )

        mock_market_book = mock.Mock()
        mock_market_book.runners = [
            mock.Mock(selection_id=1234, handicap=0, adjustment_factor=20)
        ]
        mock_market = mock.Mock(
            market_type="PLACE", blotter=[mock_order], market_book=mock_market_book
        )
        self.middleware._process_runner_removal(mock_market, 12345, 0, 50)

        # The liability of £200 is reduced by the non runner's adjustment factor of 50%
        self.assertEqual(mock_order.order_type.liability, 100)
        # Size matched should be 100 / (10.0-1.0) \approx 11.11
        self.assertEqual(11.11, mock_order.current_order.size_matched)

    def test__calculate_reduction_factor(self):
        self.assertEqual(self.middleware._calculate_reduction_factor(10, 10), 9)
        self.assertEqual(self.middleware._calculate_reduction_factor(1000, 0), 1000)
        self.assertEqual(self.middleware._calculate_reduction_factor(1000, 5), 950)
        self.assertEqual(self.middleware._calculate_reduction_factor(3.2, 75.18), 1.01)
        self.assertEqual(self.middleware._calculate_reduction_factor(10, 75.18), 2.48)
        self.assertEqual(self.middleware._calculate_reduction_factor(1.01, 75.18), 1.01)

    @mock.patch("flumine.markets.middleware.config")
    def test__process_simulated_orders_strategy_isolation(self, mock_config):
        mock_config.simulated_strategy_isolation = True
        mock_market_book = mock.Mock()
        mock_market = mock.Mock()
        mock_order = mock.Mock(
            selection_id=123, handicap=1, status=OrderStatus.EXECUTABLE, side="LAY"
        )
        mock_order.order_type.price = 1.02
        mock_order.order_type.ORDER_TYPE = OrderTypes.LIMIT
        mock_order_two = mock.Mock(
            selection_id=123, handicap=1, status=OrderStatus.PENDING
        )
        mock_order_three = mock.Mock(
            selection_id=123, handicap=1, status=OrderStatus.EXECUTABLE, simulated=False
        )
        mock_market.blotter._strategy_orders = {
            "test": [mock_order, mock_order_two, mock_order_three]
        }
        mock_market_analytics = {
            (mock_order.selection_id, mock_order.handicap): mock.Mock(traded={1: 2})
        }
        mock_market.market_book = mock_market_book
        self.middleware._process_simulated_orders(mock_market, mock_market_analytics)
        mock_order.simulated.assert_called_with(mock_market_book, {1: 2})
        mock_order_two.simulated.assert_not_called()

    @mock.patch("flumine.markets.middleware.config")
    def test__process_simulated_orders(self, mock_config):
        mock_config.simulated_strategy_isolation = False
        mock_market_book = mock.Mock()
        mock_market = mock.Mock()
        mock_order = mock.Mock(
            selection_id=123, handicap=1, status=OrderStatus.EXECUTABLE, side="LAY"
        )
        mock_order.order_type.price = 1.02
        mock_order.order_type.ORDER_TYPE = OrderTypes.LIMIT
        mock_order_two = mock.Mock(
            selection_id=123, handicap=1, status=OrderStatus.PENDING
        )
        mock_order_three = mock.Mock(
            selection_id=123, handicap=1, status=OrderStatus.EXECUTABLE, simulated=False
        )
        mock_market.blotter.live_orders = [
            mock_order,
            mock_order_two,
            mock_order_three,
        ]
        mock_market_analytics = {
            (mock_order.selection_id, mock_order.handicap): mock.Mock(traded={1: 2})
        }
        mock_market.market_book = mock_market_book
        self.middleware._process_simulated_orders(mock_market, mock_market_analytics)
        mock_order.simulated.assert_called_with(mock_market_book, {1: 2})
        mock_order_two.simulated.assert_not_called()

    def test__sort_orders(self):
        order_one = mock.Mock(side="LAY", bet_id=1)
        order_one.order_type.price = 1.01
        order_two = mock.Mock(side="LAY", bet_id=2)
        order_two.order_type.price = 1.02
        order_three = mock.Mock(side="LAY", bet_id=3)
        order_three.order_type.price = 1.01
        order_four = mock.Mock(side="BACK", bet_id=4)
        order_four.order_type.price = 1.2
        order_five = mock.Mock(side="BACK", bet_id=5)
        order_five.order_type.price = 1.2
        order_six = mock.Mock(side="BACK", bet_id=6)
        order_six.order_type.price = 1.19
        order_seven = mock.Mock(side="BACK", bet_id=6)
        order_seven.order_type.price = "ERROR"
        order_seven.order_type.ORDER_TYPE = OrderTypes.MARKET_ON_CLOSE
        orders = [
            order_one,
            order_two,
            order_three,
            order_four,
            order_five,
            order_six,
            order_seven,
        ]
        self.assertEqual(
            self.middleware._sort_orders(orders),
            [
                order_two,
                order_one,
                order_three,
                order_six,
                order_four,
                order_five,
                order_seven,
            ],
        )

    @mock.patch("flumine.markets.middleware.RunnerAnalytics")
    def test__process_runner(self, mock_runner_analytics):
        market_analytics = {}
        mock_runner = mock.Mock()
        self.middleware._process_runner(market_analytics, mock_runner)
        self.assertEqual(len(market_analytics), 1)
        self.middleware._process_runner(market_analytics, mock_runner)
        self.assertEqual(len(market_analytics), 1)
        mock_runner_analytics.assert_called_with(mock_runner)
        mock_runner_analytics().assert_called_with(mock_runner)


class RunnerAnalyticsTest(unittest.TestCase):
    def setUp(self) -> None:
        self.mock_runner = mock.Mock()
        self.mock_runner.ex.traded_volume = [{"price": 1.01, "size": 2}]
        self.runner_analytics = RunnerAnalytics(self.mock_runner)

    def test_init(self):
        self.assertEqual(self.runner_analytics._runner, self.mock_runner)
        self.assertEqual(
            self.runner_analytics._traded_volume, self.mock_runner.ex.traded_volume
        )
        self.assertEqual(self.runner_analytics.traded, {})
        self.assertEqual(self.runner_analytics._p_v, {1.01: 2})

    @mock.patch("flumine.markets.middleware.RunnerAnalytics._calculate_traded")
    def test_call(self, mock__calculate_traded):
        mock_runner = mock.Mock()
        self.runner_analytics(mock_runner)
        mock__calculate_traded.assert_called_with(mock_runner.ex.traded_volume)
        self.assertEqual(
            self.runner_analytics._traded_volume, mock_runner.ex.traded_volume
        )
        self.assertEqual(self.runner_analytics.traded, mock__calculate_traded())
        self.assertEqual(self.runner_analytics._runner, mock_runner)

    def test__calculate_traded_dict_empty(self):
        self.runner_analytics._traded_volume = []
        self.assertEqual(self.runner_analytics._calculate_traded([]), {})

    def test__calculate_traded_dict_same(self):
        traded_volume = [{"price": 1.01, "size": 69}]
        self.runner_analytics._traded_volume = [{"price": 1.01, "size": 69}]
        self.runner_analytics._p_v = {1.01: 69}
        self.assertEqual(self.runner_analytics._calculate_traded(traded_volume), {})
        self.assertEqual(self.runner_analytics._p_v, {1.01: 69})

    def test__calculate_traded_dict_new(self):
        traded_volume = [{"price": 1.01, "size": 69}]
        self.runner_analytics._traded_volume = []
        self.assertEqual(
            self.runner_analytics._calculate_traded(traded_volume), {1.01: 67.0}
        )
        self.assertEqual(self.runner_analytics._p_v, {1.01: 69})

    def test__calculate_traded_dict_new_multi(self):
        traded_volume = [
            {"price": 1.01, "size": 69},
            {"price": 10, "size": 32},
        ]
        self.runner_analytics._traded_volume = [{"price": 1.01, "size": 30}]
        self.runner_analytics._p_v = {1.01: 30}
        self.assertEqual(
            self.runner_analytics._calculate_traded(traded_volume),
            {1.01: 39.0, 10: 32},
        )
        self.assertEqual(self.runner_analytics._p_v, {1.01: 69, 10: 32})


class SimulatedSportsDataMiddlewareTest(unittest.TestCase):
    def setUp(self) -> None:
        self.middleware = SimulatedSportsDataMiddleware("cricketSubscription", "test")

    def test_init(self):
        self.assertEqual(self.middleware.operation, "cricketSubscription")
        self.assertEqual(self.middleware.directory, "test")
        self.assertIsNone(self.middleware._gen)
        self.assertIsNone(self.middleware._next)

    @mock.patch(
        "flumine.markets.middleware.call_strategy_error_handling", return_value=True
    )
    def test_call(self, mock_call_strategy_error_handling):
        mock_gen = mock.Mock()
        mock_update_one = mock.Mock(publish_time_epoch=121, market_id="1.1")
        mock_update_two = mock.Mock(publish_time_epoch=122, market_id="1.1")
        mock_update_three = mock.Mock(publish_time_epoch=123, market_id="1.1")
        mock_update_four = mock.Mock(publish_time_epoch=121, market_id="1.2")
        mock_gen.__next__ = mock.Mock(
            side_effect=[[mock_update_four], [mock_update_two], [mock_update_three]]
        )
        self.middleware._gen = mock_gen
        self.middleware._next = [mock_update_one]
        mock_strategy = mock.Mock(stream_ids=[456])
        mock_market = mock.Mock(market_id="1.1")
        mock_market.market_book.publish_time_epoch = 123
        mock_market.market_book.streaming_unique_id = 456
        mock_market.flumine.strategies = [mock_strategy]
        self.middleware(mock_market)
        mock_call_strategy_error_handling.call_count = 2
        mock_call_strategy_error_handling.assert_has_calls(
            [
                mock.call(
                    mock_strategy.check_sports_data, mock_market, mock_update_one
                ),
                mock.call(
                    mock_strategy.process_sports_data, mock_market, mock_update_one
                ),
                mock.call(
                    mock_strategy.check_sports_data, mock_market, mock_update_two
                ),
                mock.call(
                    mock_strategy.process_sports_data, mock_market, mock_update_two
                ),
            ]
        )

    @mock.patch(
        "flumine.markets.middleware.SimulatedSportsDataMiddleware._create_generator"
    )
    def test_add_market(self, mock__create_generator):
        mock_market = mock.Mock(market_id="marketId")
        self.assertIsNone(self.middleware.add_market(mock_market))
        mock__create_generator.assert_called_with(
            "test/marketId", "cricketSubscription", 123
        )
        self.assertEqual(self.middleware._gen, mock__create_generator()())
        self.assertEqual(self.middleware._next, next(mock__create_generator()()))

    def test_remove_market(self):
        self.middleware.remove_market(mock.Mock())
        self.assertIsNone(self.middleware._gen)
        self.assertIsNone(self.middleware._next)

    @mock.patch("flumine.markets.middleware.FlumineHistoricalGeneratorStream")
    @mock.patch("flumine.markets.middleware.HistoricListener")
    def test__create_generator(self, mock_listener, mock_stream):
        gen = self.middleware._create_generator(
            "test/marketId", "cricketSubscription", 123
        )
        mock_listener.assert_called_with(max_latency=None, update_clk=False)
        mock_stream.assert_called_with(
            file_path="test/marketId",
            listener=mock_listener(),
            operation="cricketSubscription",
            unique_id=123,
        )
        self.assertEqual(gen, mock_stream().get_generator())



import unittest
from unittest import mock
from betfairlightweight.resources.bettingresources import PriceSize

from flumine.order.order import OrderStatus, OrderTypes
from flumine import config
from flumine.markets.market import Market
from flumine.markets.markets import Markets
from flumine.order.order import (
    BaseOrder,
    BetfairOrder,
)
from flumine.order import process
from flumine.strategy.strategy import Strategies
from flumine.utils import create_cheap_hash


class BaseOrderTest(unittest.TestCase):
    def setUp(self) -> None:
        mock_client = mock.Mock(paper_trade=False)
        self.mock_trade = mock.Mock(client=mock_client)
        self.mock_order_type = mock.Mock()
        self.order = BaseOrder(self.mock_trade, "BACK", self.mock_order_type, 1)
        config.simulated = True

    def tearDown(self) -> None:
        config.simulated = False

    @mock.patch("flumine.order.process.process_current_order")
    def test_process_current_orders_with_default_sep(self, mock_process_current_order):
        mock_log_control = mock.Mock()
        mock_add_market = mock.Mock()
        market_book = mock.Mock()
        markets = Markets()
        market = Market(
            flumine=mock.Mock(), market_id="market_id", market_book=market_book
        )
        markets.add_market("market_id", market)
        strategies = Strategies()
        cheap_hash = create_cheap_hash("strategy_name", 13)
        trade = mock.Mock(market_id="market_id")
        trade.strategy.name_hash = cheap_hash
        current_order = mock.Mock(
            customer_order_ref=f"{cheap_hash}I123", market_id="market_id", bet_id=None
        )
        betfair_order = BetfairOrder(trade=trade, side="BACK", order_type=mock.Mock())
        betfair_order.id = "123"
        betfair_order.complete = True
        market.blotter["123"] = betfair_order
        event = mock.Mock(event=[mock.Mock(orders=[current_order])])

        process.process_current_orders(
            markets=markets,
            strategies=strategies,
            event=event,
            log_control=mock_log_control,
            add_market=mock_add_market,
        )
        mock_process_current_order.assert_called_with(
            betfair_order, current_order, mock_log_control
        )
        self.assertEqual(market.blotter._live_orders, [])

    def test_process_current_order(self):
        mock_order = mock.Mock(status=OrderStatus.EXECUTABLE)
        mock_order.current_order.status = "EXECUTION_COMPLETE"
        mock_current_order = mock.Mock()
        mock_log_control = mock.Mock()
        process.process_current_order(mock_order, mock_current_order, mock_log_control)
        mock_order.update_current_order.assert_called_with(mock_current_order)
        mock_order.execution_complete.assert_called()

    @mock.patch("flumine.order.process.OrderEvent")
    def test_process_current_order_async(self, mock_order_event):
        mock_order = mock.Mock(status=OrderStatus.EXECUTABLE, async_=True, bet_id=None)
        mock_order.current_order.status = "EXECUTION_COMPLETE"
        mock_current_order = mock.Mock(bet_id=1234)
        mock_log_control = mock.Mock()
        process.process_current_order(mock_order, mock_current_order, mock_log_control)
        mock_order.update_current_order.assert_called_with(mock_current_order)
        mock_order.execution_complete.assert_called()
        self.assertEqual(mock_order.bet_id, 1234)
        mock_order.responses.placed.assert_called_with()
        mock_order_event.assert_called_with(mock_order)
        mock_log_control.assert_called_with(mock_order_event())

    def test_create_order_from_current(self):
        mock_add_market = mock.Mock()
        market_book = mock.Mock()
        mock_client = mock.Mock()
        markets = Markets()
        market = Market(
            flumine=mock.Mock(), market_id="market_id", market_book=market_book
        )
        markets.add_market("market_id", market)
        cheap_hash = create_cheap_hash("strategy_name", 13)
        strategy = mock.Mock(name_hash=cheap_hash)
        strategies = Strategies()
        strategies(strategy=strategy, clients=mock.Mock())
        current_order = mock.Mock(
            customer_order_ref=f"{cheap_hash}I123",
            market_id="market_id",
            bet_id=None,
            selection_id="selection_id",
            handicap="handicap",
            order_type="LIMIT",
            price_size=PriceSize(price=10.0, size=2.0),
            persistence_type="LAPSE",
        )

        new_order = process.create_order_from_current(
            markets=markets,
            strategies=strategies,
            current_order=current_order,
            add_market=mock_add_market,
            client=mock_client,
        )
        self.assertEqual(market.blotter["123"], new_order)
        self.assertEqual(new_order.market_id, "market_id")
        self.assertEqual(new_order.selection_id, "selection_id")
        self.assertEqual(new_order.handicap, "handicap")
        self.assertEqual(new_order.order_type.ORDER_TYPE, OrderTypes.LIMIT)
        self.assertEqual(new_order.order_type.size, 2.0)
        self.assertEqual(new_order.order_type.price, 10.0)
        self.assertEqual(new_order.client, mock_client)



import datetime
import logging
from typing import Optional
from collections import defaultdict
from betfairlightweight.resources.bettingresources import MarketBook, MarketCatalogue

from .. import config
from .blotter import Blotter
from ..execution.transaction import Transaction

logger = logging.getLogger(__name__)


class Market:
    """
    Market data structure to hold latest marketBook,
    marketCatalogue and various properties. Also
    allows order placement through the Transaction
    class.
    """

    def __init__(
        self,
        flumine,
        market_id: str,
        market_book: MarketBook,
        market_catalogue: MarketCatalogue = None,
    ):
        self.flumine = flumine
        self.market_id = market_id
        self.closed = False
        self.date_time_created = datetime.datetime.utcnow()
        self.date_time_closed = None
        self.market_book = market_book
        self.market_catalogue = market_catalogue
        self.update_market_catalogue = True
        self.orders_cleared = []
        self.market_cleared = []
        self.context = {"simulated": {}}  # data store (raceCard / scores etc)
        self.blotter = Blotter(market_id)
        self._transaction_id = 0

    def __call__(self, market_book: MarketBook):
        if self.market_book and market_book.version != self.market_book.version:
            self.update_market_catalogue = True
        self.market_book = market_book

    def open_market(self) -> None:
        self.closed = False
        self.orders_cleared = []
        self.market_cleared = []
        logger.info(
            "Market {0} opened".format(self.market_id),
            extra=self.info,
        )

    def close_market(self) -> None:
        self.closed = True
        self.date_time_closed = datetime.datetime.utcnow()
        logger.info(
            "Market {0} closed".format(self.market_id),
            extra=self.info,
        )

    def transaction(self, async_place_orders: bool = None, client=None) -> Transaction:
        if async_place_orders is None:
            async_place_orders = config.async_place_orders
        if client is None:
            client = self.flumine.clients.get_default()
        self._transaction_id += 1
        return Transaction(
            self,
            id_=self._transaction_id,
            async_place_orders=async_place_orders,
            client=client,
        )

    # order
    def place_order(
        self,
        order,
        market_version: int = None,
        execute: bool = True,
        force: bool = False,
        client=None,
    ) -> bool:
        with self.transaction(client=client) as t:
            return t.place_order(order, market_version, execute, force)

    def cancel_order(
        self, order, size_reduction: float = None, force: bool = False
    ) -> bool:
        with self.transaction(client=order.client) as t:
            return t.cancel_order(order, size_reduction, force)

    def update_order(
        self, order, new_persistence_type: str, force: bool = False
    ) -> bool:
        with self.transaction(client=order.client) as t:
            return t.update_order(order, new_persistence_type, force)

    def replace_order(
        self, order, new_price: float, market_version: int = None, force: bool = False
    ) -> bool:
        with self.transaction(client=order.client) as t:
            return t.replace_order(order, new_price, market_version, force)

    @property
    def event(self) -> dict:
        event = defaultdict(list)
        for market in self.flumine.markets:
            if (
                self.event_id == market.event_id
                and self.market_start_datetime == market.market_start_datetime
            ):
                event[market.market_type].append(market)
        return event

    @property
    def event_type_id(self) -> str:
        if self.market_catalogue:
            return self.market_catalogue.event_type.id
        elif self.market_book:
            return self.market_book.market_definition.event_type_id

    @property
    def event_id(self) -> str:
        if self.market_catalogue:
            return self.market_catalogue.event.id
        elif self.market_book:
            return self.market_book.market_definition.event_id

    @property
    def market_type(self) -> str:
        if self.market_catalogue:
            return self.market_catalogue.description.market_type
        elif self.market_book:
            return self.market_book.market_definition.market_type

    @property
    def seconds_to_start(self) -> float:
        return (self.market_start_datetime - datetime.datetime.utcnow()).total_seconds()

    @property
    def elapsed_seconds_closed(self) -> Optional[float]:
        if self.closed and self.date_time_closed:
            return (datetime.datetime.utcnow() - self.date_time_closed).total_seconds()

    @property
    def market_start_datetime(self):
        if self.market_book:
            return self.market_book.market_definition.market_time
        elif self.market_catalogue:
            return self.market_catalogue.market_start_time
        else:
            return datetime.datetime.utcfromtimestamp(0)

    @property
    def market_start_hour_minute(self) -> Optional[str]:
        msd = self.market_start_datetime
        if msd:
            return "{0:02d}{1:02d}".format(msd.hour, msd.minute)

    @property
    def event_name(self) -> Optional[str]:
        if self.market_catalogue:
            return self.market_catalogue.event.name
        elif self.market_book:
            return self.market_book.market_definition.event_name

    @property
    def country_code(self) -> Optional[str]:
        if self.market_catalogue:
            return self.market_catalogue.event.country_code
        elif self.market_book:
            return self.market_book.market_definition.country_code

    @property
    def venue(self) -> Optional[str]:
        if self.market_catalogue:
            return self.market_catalogue.event.venue
        elif self.market_book:
            return self.market_book.market_definition.venue

    @property
    def race_type(self) -> Optional[str]:
        if self.market_catalogue:
            return self.market_catalogue.description.race_type
        elif self.market_book:
            return self.market_book.market_definition.race_type

    @property
    def status(self) -> Optional[str]:
        if self.market_book:
            return self.market_book.status

    def cleared(self, client) -> dict:
        orders = self.blotter.client_orders(client, matched_only=True)
        profit = round(sum([order.profit for order in orders]), 2)
        return {
            "marketId": self.market_id,
            "eventId": self.event_id,
            "eventTypeId": self.event_type_id,
            "customerStrategyRef": config.customer_strategy_ref,
            "lastMatchedDate": None,
            "placedDate": None,
            "settledDate": None,
            "betCount": len(orders),
            "betOutcome": "WON" if profit >= 0 else "LOST",
            "commission": round(max(profit * client.commission_base, 0), 2),
            "profit": profit,
        }

    @property
    def info(self) -> dict:
        return {
            "market_id": self.market_id,
            "event_id": self.event_id,
            "event_type_id": self.event_type_id,
            "event_name": self.event_name,
            "market_type": self.market_type,
            "market_start_datetime": str(self.market_start_datetime),
            "country_code": self.country_code,
            "venue": self.venue,
            "race_type": self.race_type,
            "orders_cleared": self.orders_cleared,
            "market_cleared": self.market_cleared,
            "closed": self.closed,
        }



import unittest

from flumine import FlumineSimulation, clients, BaseStrategy, config
from flumine.order.trade import Trade
from flumine.order.order import OrderStatus
from flumine.order.ordertype import LimitOrder, MarketOnCloseOrder, LimitOnCloseOrder
from flumine.utils import get_price, get_nearest_price


class IntegrationTest(unittest.TestCase):
    def setUp(self) -> None:
        # change config to raise errors
        config.raise_errors = True

    def test_simulation_basic(self):
        class Ex(BaseStrategy):
            def check_market_book(self, market, market_book):
                return True

            def process_market_book(self, market, market_book):
                return

        client = clients.SimulatedClient()
        framework = FlumineSimulation(client=client)
        strategy = Ex(market_filter={"markets": ["tests/resources/BASIC-1.132153978"]})
        framework.add_strategy(strategy)
        framework.run()

    def test_simulation_pro(self):
        class LimitOrders(BaseStrategy):
            def check_market_book(self, market, market_book):
                if not market_book.inplay and market.seconds_to_start < 100:
                    return True

            def process_market_book(self, market, market_book):
                with market.transaction() as t:
                    for runner in market_book.runners:
                        if runner.status == "ACTIVE":
                            back = get_price(runner.ex.available_to_back, 0)
                            runner_context = self.get_runner_context(
                                market.market_id, runner.selection_id
                            )
                            if runner_context.trade_count == 0:
                                trade = Trade(
                                    market_book.market_id,
                                    runner.selection_id,
                                    runner.handicap,
                                    self,
                                )
                                order = trade.create_order(
                                    side="BACK",
                                    order_type=LimitOrder(back, 2.00),
                                )
                                t.place_order(order)

        class LimitReplaceOrders(BaseStrategy):
            def check_market_book(self, market, market_book):
                if not market_book.inplay and market.seconds_to_start < 100:
                    return True

            def process_market_book(self, market, market_book):
                with market.transaction() as t:
                    for runner in market_book.runners:
                        if runner.status == "ACTIVE":
                            runner_context = self.get_runner_context(
                                market.market_id, runner.selection_id
                            )
                            if runner_context.trade_count == 0:
                                trade = Trade(
                                    market_book.market_id,
                                    runner.selection_id,
                                    runner.handicap,
                                    self,
                                )
                                order = trade.create_order(
                                    side="BACK",
                                    order_type=LimitOrder(1000, 2.00),
                                )
                                t.place_order(order)

            def process_orders(self, market, orders: list) -> None:
                with market.transaction() as t:
                    for order in orders:
                        if order.status == OrderStatus.EXECUTABLE:
                            if order.size_matched == 0:
                                t.replace_order(order, new_price=1.01)

        class LimitOrdersInplay(BaseStrategy):
            def check_market_book(self, market, market_book):
                if market_book.inplay:
                    return True

            def process_market_book(self, market, market_book):
                for runner in market_book.runners:
                    if runner.status == "ACTIVE" and runner.last_price_traded < 2:
                        lay = get_price(runner.ex.available_to_lay, 0)
                        trade = Trade(
                            market_book.market_id,
                            runner.selection_id,
                            runner.handicap,
                            self,
                        )
                        order = trade.create_order(
                            side="LAY",
                            order_type=LimitOrder(lay, 2.00),
                        )
                        market.place_order(order)

            def process_orders(self, market, orders):
                for order in orders:
                    if order.status == OrderStatus.EXECUTABLE:
                        if order.elapsed_seconds and order.elapsed_seconds > 2:
                            market.cancel_order(order)

        class LimitOnCloseOrders(BaseStrategy):
            def check_market_book(self, market, market_book):
                if not market_book.inplay and market.seconds_to_start < 100:
                    return True

            def process_market_book(self, market, market_book):
                for runner in market_book.runners:
                    if runner.status == "ACTIVE":
                        runner_context = self.get_runner_context(
                            market.market_id, runner.selection_id
                        )
                        if runner_context.trade_count == 0:
                            back = get_price(runner.ex.available_to_back, 0)
                            trade = Trade(
                                market_book.market_id,
                                runner.selection_id,
                                runner.handicap,
                                self,
                            )
                            order = trade.create_order(
                                side="BACK",
                                order_type=LimitOnCloseOrder(liability=10, price=back),
                            )
                            market.place_order(order)

        class MarketOnCloseOrders(BaseStrategy):
            def check_market_book(self, market, market_book):
                if not market_book.inplay and market.seconds_to_start < 100:
                    return True

            def process_market_book(self, market, market_book):
                for runner in market_book.runners:
                    if runner.status == "ACTIVE":
                        runner_context = self.get_runner_context(
                            market.market_id, runner.selection_id
                        )
                        if runner_context.trade_count == 0:
                            trade = Trade(
                                market_book.market_id,
                                runner.selection_id,
                                runner.handicap,
                                self,
                            )
                            order = trade.create_order(
                                side="LAY",
                                order_type=MarketOnCloseOrder(100.00),
                            )
                            market.place_order(order)

        client = clients.SimulatedClient()
        framework = FlumineSimulation(client=client)
        limit_strategy = LimitOrders(
            market_filter={"markets": ["tests/resources/PRO-1.170258213"]},
            max_order_exposure=1000,
            max_selection_exposure=105,
            max_trade_count=1,
        )
        framework.add_strategy(limit_strategy)
        limit_replace_strategy = LimitReplaceOrders(
            market_filter={"markets": ["tests/resources/PRO-1.170258213"]},
            max_order_exposure=1000,
            max_selection_exposure=105,
            max_trade_count=1,
        )
        framework.add_strategy(limit_replace_strategy)
        limit_inplay_strategy = LimitOrdersInplay(
            market_filter={"markets": ["tests/resources/PRO-1.170258213"]},
            max_order_exposure=1000,
            max_selection_exposure=105,
        )
        framework.add_strategy(limit_inplay_strategy)
        limit_on_close_strategy = LimitOnCloseOrders(
            market_filter={"markets": ["tests/resources/PRO-1.170258213"]},
            max_order_exposure=1000,
            max_selection_exposure=105,
        )
        framework.add_strategy(limit_on_close_strategy)
        market_strategy = MarketOnCloseOrders(
            market_filter={"markets": ["tests/resources/PRO-1.170258213"]},
            max_order_exposure=1000,
            max_selection_exposure=105,
        )
        framework.add_strategy(market_strategy)
        framework.run()

        self.assertEqual(len(framework.markets), 1)

        for market in framework.markets:
            limit_orders = market.blotter.strategy_orders(limit_strategy)
            self.assertEqual(round(sum([o.profit for o in limit_orders]), 2), -16.8)
            self.assertEqual(len(limit_orders), 14)
            limit_replace_orders = market.blotter.strategy_orders(
                limit_replace_strategy
            )
            self.assertEqual(
                round(sum([o.profit for o in limit_replace_orders]), 2), -16.8
            )
            self.assertEqual(len(limit_replace_orders), 28)
            limit_inplay_orders = market.blotter.strategy_orders(limit_inplay_strategy)
            self.assertEqual(
                round(sum([o.profit for o in limit_inplay_orders]), 2), 19.88
            )
            self.assertEqual(len(limit_inplay_orders), 14)
            limit_on_close_orders = market.blotter.strategy_orders(
                limit_on_close_strategy
            )
            self.assertEqual(
                round(sum([o.profit for o in limit_on_close_orders]), 2), -53.20
            )
            self.assertEqual(len(limit_on_close_orders), 14)
            market_orders = market.blotter.strategy_orders(market_strategy)
            self.assertEqual(round(sum([o.profit for o in market_orders]), 2), -6.68)
            self.assertEqual(len(market_orders), 14)
            # check transaction count
            self.assertEqual(market._transaction_id, 5196)

    def test_simulation_multi_clients(self):
        class LimitOrders(BaseStrategy):
            def check_market_book(self, market, market_book):
                if market_book.inplay:
                    return True

            def process_market_book(self, market, market_book):
                with market.transaction(client=self.context["client"]) as t:
                    for runner in market_book.runners:
                        if runner.status == "ACTIVE":
                            back = get_price(runner.ex.available_to_back, 0)
                            runner_context = self.get_runner_context(
                                market.market_id, runner.selection_id
                            )
                            if runner_context.trade_count == 0:
                                trade = Trade(
                                    market_book.market_id,
                                    runner.selection_id,
                                    runner.handicap,
                                    self,
                                )
                                order = trade.create_order(
                                    side="BACK",
                                    order_type=LimitOrder(back, 2.00),
                                )
                                t.place_order(order)

            def process_orders(self, market, orders):
                for order in orders:
                    if order.status == OrderStatus.EXECUTABLE:
                        if order.elapsed_seconds and order.elapsed_seconds > 2:
                            new_price = get_nearest_price(order.order_type.price - 1)
                            market.replace_order(order, new_price)
                        elif order.elapsed_seconds and order.elapsed_seconds > 10:
                            market.cancel_order(order)

        client_bpe_on = clients.SimulatedClient()
        client_bpe_off = clients.SimulatedClient(best_price_execution=False)
        framework = FlumineSimulation()
        framework.add_client(client_bpe_on)
        framework.add_client(client_bpe_off)
        limit_strategy_bpe_on = LimitOrders(
            market_filter={"markets": ["tests/resources/PRO-1.170258213"]},
            max_order_exposure=1000,
            max_selection_exposure=105,
            max_trade_count=100,
            context={"client": client_bpe_on},
        )
        framework.add_strategy(limit_strategy_bpe_on)
        limit_strategy_bpe_off = LimitOrders(
            market_filter={"markets": ["tests/resources/PRO-1.170258213"]},
            max_order_exposure=1000,
            max_selection_exposure=105,
            max_trade_count=100,
            context={"client": client_bpe_off},
        )
        framework.add_strategy(limit_strategy_bpe_off)
        framework.run()

        self.assertEqual(len(framework.markets), 1)

        for market in framework.markets:
            limit_orders_bpe_on = market.blotter.strategy_orders(limit_strategy_bpe_on)
            self.assertEqual(
                limit_orders_bpe_on, market.blotter.client_orders(client_bpe_on)
            )
            self.assertEqual(
                round(sum([o.profit for o in limit_orders_bpe_on]), 2), -17.75
            )
            self.assertEqual(len(limit_orders_bpe_on), 15)
            limit_orders_bpe_off = market.blotter.strategy_orders(
                limit_strategy_bpe_off
            )
            self.assertEqual(
                limit_orders_bpe_off, market.blotter.client_orders(client_bpe_off)
            )
            self.assertEqual(
                round(sum([o.profit for o in limit_orders_bpe_off]), 2),
                -19.75,
            )
            self.assertEqual(len(limit_orders_bpe_off), 14)

    def test_event_processing(self):
        client = clients.SimulatedClient()
        framework = FlumineSimulation(client=client)

        class LimitOrdersInplay(BaseStrategy):
            def check_market_book(self, market, market_book):
                if market_book.inplay:
                    return True

            def process_market_book(self, market, market_book):
                for runner in market_book.runners:
                    if runner.status == "ACTIVE" and runner.last_price_traded < 2:
                        lay = get_price(runner.ex.available_to_lay, 0)
                        trade = Trade(
                            market_book.market_id,
                            runner.selection_id,
                            runner.handicap,
                            self,
                        )
                        order = trade.create_order(
                            side="LAY",
                            order_type=LimitOrder(lay, 2.00),
                        )
                        market.place_order(order)

            def process_orders(self, market, orders):
                for order in orders:
                    if order.status == OrderStatus.EXECUTABLE:
                        if order.elapsed_seconds and order.elapsed_seconds > 2:
                            market.cancel_order(order)

        limit_inplay_strategy = LimitOrdersInplay(
            market_filter={
                "markets": [
                    "tests/resources/SELF-1.181223994",
                    "tests/resources/SELF-1.181223995",
                    "tests/resources/PRO-1.170258213",
                ],
                "event_processing": True,
            },
            max_order_exposure=1000,
            max_selection_exposure=105,
        )
        framework.add_strategy(limit_inplay_strategy)
        framework.run()

        self.assertEqual(len(framework.markets), 3)

        # Different event
        win_market = framework.markets.markets["1.170258213"]
        limit_inplay_orders = win_market.blotter.strategy_orders(limit_inplay_strategy)
        self.assertEqual(round(sum([o.profit for o in limit_inplay_orders]), 2), 19.88)
        self.assertEqual(len(limit_inplay_orders), 14)
        self.assertEqual(win_market._transaction_id, 165)

        # Same event
        win_market = framework.markets.markets["1.181223994"]
        limit_inplay_orders = win_market.blotter.strategy_orders(limit_inplay_strategy)
        self.assertEqual(round(sum([o.profit for o in limit_inplay_orders]), 2), 101.44)
        self.assertEqual(len(limit_inplay_orders), 86)
        self.assertEqual(win_market._transaction_id, 1329)

        place_market = framework.markets.markets["1.181223995"]
        limit_inplay_orders = place_market.blotter.strategy_orders(
            limit_inplay_strategy
        )
        self.assertEqual(round(sum([o.profit for o in limit_inplay_orders]), 2), -95.02)
        self.assertEqual(len(limit_inplay_orders), 200)
        self.assertEqual(place_market._transaction_id, 2436)

    def tearDown(self) -> None:
        config.simulated = False
        config.raise_errors = False



import time
import queue
import logging
from betfairlightweight import BetfairError
from tenacity import retry

from .basestream import BaseStream
from ..events.events import SportsDataEvent

logger = logging.getLogger(__name__)

RETRY_WAIT = BaseStream.RETRY_WAIT


class SportsDataStream(BaseStream):
    @retry(wait=RETRY_WAIT)
    def run(self) -> None:
        time.sleep(2)  # 2s delay to wait for market streams to start
        logger.info(
            "Starting SportsDataStream {0}".format(self.stream_id),
            extra={
                "stream_id": self.stream_id,
                "sports_data_filter": self.sports_data_filter,
                "streaming_timeout": self.streaming_timeout,
            },
        )
        if not self._output_thread.is_alive():
            logger.info(
                "Starting output_thread (SportsDataStream {0})".format(self.stream_id)
            )
            self._output_thread.start()

        self._stream = self.betting_client.streaming.create_stream(
            unique_id=self.stream_id, host="sports_data", listener=self._listener
        )
        try:
            if self.sports_data_filter == "raceSubscription":
                self.stream_id = self._stream.subscribe_to_races()
            elif self.sports_data_filter == "cricketSubscription":
                self.stream_id = self._stream.subscribe_to_cricket_matches()
            else:
                raise ValueError("Unknown sports data filter")
            self._stream.start()
        except BetfairError:
            logger.error(
                "SportsDataStream {0} run error".format(self.stream_id), exc_info=True
            )
            raise
        except Exception:
            logger.critical(
                "SportsDataStream {0} run error".format(self.stream_id), exc_info=True
            )
            raise
        logger.info("Stopped SportsDataStream {0}".format(self.stream_id))

    def handle_output(self) -> None:
        """Handles output from stream."""
        while self.is_alive():
            try:
                sports_data = self._output_queue.get(
                    block=True, timeout=self.streaming_timeout
                )
            except queue.Empty:
                sports_data = self._listener.snap(
                    market_ids=self.flumine.markets.open_market_ids
                )
            if sports_data:
                self.flumine.handler_queue.put(SportsDataEvent(sports_data))

        logger.info(
            "Stopped output_thread (SportsDataStream {0})".format(self.stream_id)
        )



import unittest
from unittest import mock
from queue import Queue

from flumine.controls.loggingcontrols import LoggingControl, EventType


class TestLoggingControl(unittest.TestCase):
    def setUp(self):
        self.logging_control = LoggingControl()

    def test_init(self):
        self.assertIsInstance(self.logging_control.logging_queue, Queue)
        self.assertEqual(self.logging_control.cache, [])
        self.assertEqual(self.logging_control.NAME, "LOGGING_CONTROL")

    def test_run(self):
        self.logging_control.logging_queue.put(None)
        self.logging_control.run()

    def test_run_error(self):
        self.logging_control.logging_queue.put(1)
        self.logging_control.logging_queue.put(None)
        self.logging_control.run()

    @mock.patch("flumine.controls.loggingcontrols.LoggingControl._process_config")
    def test_process_event_config(self, mock_process_config):
        mock_event = mock.Mock()
        mock_event.EVENT_TYPE = EventType.CONFIG
        self.logging_control.process_event(mock_event)
        mock_process_config.assert_called_with(mock_event)

    @mock.patch("flumine.controls.loggingcontrols.LoggingControl._process_strategy")
    def test_process_event_strategy(self, mock_process_strategy):
        mock_event = mock.Mock()
        mock_event.EVENT_TYPE = EventType.STRATEGY
        self.logging_control.process_event(mock_event)
        mock_process_strategy.assert_called_with(mock_event)

    @mock.patch("flumine.controls.loggingcontrols.LoggingControl._process_market")
    def test_process_event_market(self, _process_market):
        mock_event = mock.Mock()
        mock_event.EVENT_TYPE = EventType.MARKET
        self.logging_control.process_event(mock_event)
        _process_market.assert_called_with(mock_event)

    @mock.patch("flumine.controls.loggingcontrols.LoggingControl._process_trade")
    def test_process_event_trade(self, _process_trade):
        mock_event = mock.Mock()
        mock_event.EVENT_TYPE = EventType.TRADE
        self.logging_control.process_event(mock_event)
        _process_trade.assert_called_with(mock_event)

    @mock.patch("flumine.controls.loggingcontrols.LoggingControl._process_order")
    def test_process_event_order(self, _process_order):
        mock_event = mock.Mock()
        mock_event.EVENT_TYPE = EventType.ORDER
        self.logging_control.process_event(mock_event)
        _process_order.assert_called_with(mock_event)

    @mock.patch("flumine.controls.loggingcontrols.LoggingControl._process_balance")
    def test_process_event_balance(self, _process_balance):
        mock_event = mock.Mock()
        mock_event.EVENT_TYPE = EventType.BALANCE
        self.logging_control.process_event(mock_event)
        _process_balance.assert_called_with(mock_event)

    @mock.patch(
        "flumine.controls.loggingcontrols.LoggingControl._process_cleared_orders"
    )
    def test_process_event_cleared(self, _process_cleared_orders):
        mock_event = mock.Mock()
        mock_event.EVENT_TYPE = EventType.CLEARED_ORDERS
        self.logging_control.process_event(mock_event)
        _process_cleared_orders.assert_called_with(mock_event)

    @mock.patch(
        "flumine.controls.loggingcontrols.LoggingControl._process_cleared_orders_meta"
    )
    def test_process_event_cleared_orders_meta(self, _process_cleared_orders_meta):
        mock_event = mock.Mock()
        mock_event.EVENT_TYPE = EventType.CLEARED_ORDERS_META
        self.logging_control.process_event(mock_event)
        _process_cleared_orders_meta.assert_called_with(mock_event)

    @mock.patch(
        "flumine.controls.loggingcontrols.LoggingControl._process_cleared_markets"
    )
    def test_process_event_cleared_markets(self, _process_cleared_markets):
        mock_event = mock.Mock()
        mock_event.EVENT_TYPE = EventType.CLEARED_MARKETS
        self.logging_control.process_event(mock_event)
        _process_cleared_markets.assert_called_with(mock_event)

    @mock.patch(
        "flumine.controls.loggingcontrols.LoggingControl._process_closed_market"
    )
    def test_process_event_closed(self, _closed_market):
        mock_event = mock.Mock()
        mock_event.EVENT_TYPE = EventType.CLOSE_MARKET
        self.logging_control.process_event(mock_event)
        _closed_market.assert_called_with(mock_event)

    @mock.patch("flumine.controls.loggingcontrols.LoggingControl._process_custom_event")
    def test_process_event_custom_event(self, _closed_market):
        mock_event = mock.Mock()
        mock_event.EVENT_TYPE = EventType.CUSTOM_EVENT
        self.logging_control.process_event(mock_event)
        _closed_market.assert_called_with(mock_event)

    @mock.patch("flumine.controls.loggingcontrols.LoggingControl._process_end_flumine")
    def test_process_event_end(self, _end_flumine):
        mock_event = mock.Mock()
        mock_event.EVENT_TYPE = EventType.TERMINATOR
        self.logging_control.process_event(mock_event)
        _end_flumine.assert_called_with(mock_event)
        self.assertIsNone(self.logging_control.logging_queue.get())

    def test_process_config(self):
        self.logging_control._process_config(None)

    def test_process_strategy(self):
        self.logging_control._process_strategy(None)

    def test_process_market(self):
        self.logging_control._process_market(None)

    def test_process_trade(self):
        self.logging_control._process_trade(None)

    def test_process_order(self):
        self.logging_control._process_order(None)

    def test_process_balance(self):
        self.logging_control._process_balance(None)

    def test_process_cleared_orders(self):
        self.logging_control._process_cleared_orders(None)

    def test_process_cleared_orders_meta(self):
        self.logging_control._process_cleared_orders_meta(None)

    def test_process_cleared_markets(self):
        self.logging_control._process_cleared_markets(None)

    def test_process_closed_market(self):
        self.logging_control._process_closed_market(None)

    def test_process_custom_event(self):
        self.logging_control._process_custom_event(None)

    def test_process_end_flumine(self):
        self.logging_control._process_end_flumine(None)
